const { dateRange } = require('../Utils/Utility');
const { logger } = require('../logger');

const CaseServices = require('../Services/CreateServices')("Case");
const UserServices = require('../Services/CreateServices')("User");
const ApprovalRequestServices = require('../Services/CreateServices')("ApprovalRequest");

module.exports.getDashboardCount = async (req, res) => {
    try {

        let responseCounts = {
            PENDING: { count: 0 },
            APPROVED: { count: 0 },
            REJECTED: { count: 0 }
        };

        const requestResponse = await ApprovalRequestServices.aggregate([
            { $group: { _id: "$status", count: { $sum: 1 } } },
        ]);
        if (!requestResponse.status || requestResponse.error) return res.status(500).send(requestResponse);

        const statusCounts = requestResponse.data.map((st, id) => {
            responseCounts[st._id].count = st.count;
        });

        return res.status(200).send({ status: true, counts: responseCounts });
    }
    catch (e) {
        logger.error("🚀 ~ file: Payment.controller.js:48 ~ module.exports.getAllPayments= ~ e:" + e.toString())
        return res.status(500).send({ status: false, error: e.toString() });
    }
};

module.exports.getAll = async (req, res) => {
    try {
        const user = req.local.user;
        const { status, start, end } = req.query;
        let query = {};

        if (["ADMIN", "ORG_ADMIN", "OWNER", "VALUER", "FRONTDESK"].indexOf(user?.role) < 0) {
            query = { userId: user._id }
        }
        if (status) { query.status = status };
        if (!!start && !!end) { query.createdAt = dateRange(start, end) }

        const createRequestRes = await ApprovalRequestServices.find({ query, populate: ['caseId'], start: req.local.startIndex, limit: req.local.limit });
        if (!createRequestRes.status || createRequestRes.error) return res.status(500).send(createRequestRes);

        return res.status(200).send(createRequestRes);
    } catch (e) {
        logger.error("🚀 ~ file: ApprovalRequest.controller.js:11 ~ module.exports.createTicket= ~ e:" + e.toString())
        return res.status(200).send({ status: false, error: e.toString() });
    }
}

module.exports.get = async (req, res) => {
    try {
        const id = req.params.id;

        const createTicketRes = await ApprovalRequestServices.find({ query: { _id: id }, populate: ["caseId"] });
        if (!createTicketRes.status || createTicketRes.error) return res.status(500).send(createTicketRes);
        if (createTicketRes.data.length == 0) return res.status(404).send({ status: false, error: "No Ticket Found!" })
        return res.status(200).send({ status: true, data: createTicketRes.data });
    } catch (e) {
        logger.error("🚀 ~ file: ApprovalRequest.controller.js:25 ~ module.exports.createTicket= ~ e:" + e.toString())
        return res.status(200).send({ status: false, error: e.toString() });
    }
}

module.exports.create = async (req, res) => {
    try {
        const user = req.local.user;
        let data = req.local.data;

        data.organizationId = user.organization._id;
        data.requestedBy = user;

        const hasRequestResult = await ApprovalRequestServices.find({ caseId: data.caseId });
        if (!hasRequestResult.status || hasRequestResult.error) return res.status(500).send(hasRequestResult);
        if (hasRequestResult.count > 0) return res.status(501).send({ status: false, error: `Already Requested for ${hasRequestResult.data[0].type.replace(/_/g, ' ')}` });

        const requestResult = await ApprovalRequestServices.create(data);
        if (!requestResult.status || requestResult.error) return res.status(500).send(requestResult);

        const case_data = { $push: { 'remarks': { note: data.remark, taskId: data.taskId, workType: data.workType, createdBy: user, createdAt: new Date(), } } };
        case_data.updatedBy = user._id;

        const case_res = await CaseServices.updateById(data.caseId, case_data);

        return res.status(200).send({ status: true, data: requestResult.data });

    } catch (e) {
        console.log("🚀 ~ file: ApprovalRequest.controller.js:84 ~ module.exports.createTicket= ~ e:", e)
        return res.status(200).send({ status: false, error: e });
    }
}

module.exports.update = async (req, res) => {
    try {
        const user = req.local.user;
        const { id, status } = req.params;
        let data = req.local.data;
        data.actionedBy = user;

        const hasRequestResult = await ApprovalRequestServices.findById(id);
        if (!hasRequestResult.status || hasRequestResult.error) return res.status(500).send(hasRequestResult);
        if (!hasRequestResult.data._id) return res.status(501).send({ status: false, error: 'No Request found.' });

        if (!!status && status == "ACCEPT") {
            switch (hasRequestResult.data.type) {
                case "CONVERT_TO_CASE":

                    const caseResult = await CaseServices.findById(hasRequestResult.data.caseId);
                    if (caseResult.data && !!caseResult.data.createdBy) {
                        const pickupPerson = await UserServices.findById(caseResult.Leads.createdBy);
                        if (!pickupPerson.status || pickupPerson.error) return res.status(500).send(pickupPerson);
                        datas.staff = { pickupPerson: pickupPerson.data };
                    }

                    const LeadUpdateResult = await CaseServices.updateById(hasRequestResult.data.caseId, { ...datas, type: "CASE", caseStatus: "NEW" });
                    if (!LeadUpdateResult.status || LeadUpdateResult.error) return res.status(500).send(LeadUpdateResult);

                    break;
                case "CANCEL_CASE":
                    const CaseUpdateResult = await CaseServices.updateById(hasRequestResult.data.caseId, { caseStatus: "CANCELLED" });
                    if (!CaseUpdateResult.status || CaseUpdateResult.error) return res.status(500).send(CaseUpdateResult);

                    break;
                default:
                    break;
            }

            const updateTicketRes = await ApprovalRequestServices.updateById(id, { ...data, status: "APPROVED" });
            if (!updateTicketRes.status || updateTicketRes.error) return res.status(500).send(updateTicketRes);

            return res.status(200).send(updateTicketRes);
        }
        else {
            const updateTicketRes = await ApprovalRequestServices.updateById(id, { ...data, status: 'REJECTED' });
            if (!updateTicketRes.status || updateTicketRes.error) return res.status(500).send(updateTicketRes);

            return res.status(200).send(updateTicketRes);
        }
    }
    catch (e) {
        console.log("🚀 ~ file: ApprovalRequest.controller.js:100 ~ module.exports.updateTicket= ~ e:", e)
        return res.status(200).send({ status: false, error: e });
    }
}