// const BranchServices = require('../Services/Branch.services');
const BranchServices = require('../Services/CreateServices')('Branch');
const ClientServices = require('../Services/CreateServices')('Clients');
const CaseServices = require('../Services/Case.services');
const UserServices = require('../Services/User.services');
const TransferBranchHistoryServices = require('../Services/TransferBranchHistory.services');
const OrganizationServices = require('../Services/Organization.services');
const ZoneServices = require('../Services/Zone.services');

const CreditOfficerServices = require('../Services/CreateServices')('CreditOfficer');
const { logger } = require('../logger');

/**
 * @description This function is used for get all branches from the database. 
 * @param {*} req 
 * @param {*} res 
 * @returns Object BranchResult { status : true, branches : array }
 * - { status: false, error: string }
 */
module.exports.getAllBranches = async (req, res) => {
    try {
        const { withCount = "false", searchValue, client, branch, zone } = req.query;
        let query = {};

        var count = 0;
        // query['organization._id'] = req.local.user.organization._id;
        query['status'] = true;

        if (searchValue) {
            query["$or"] = [
                { address: { $regex: searchValue, $options: 'i' } },
                { name: { $regex: searchValue, $options: 'i' } },
                { mobileNo: { $regex: searchValue, $options: 'i' } },
                { "client.name": { $regex: searchValue, $options: 'i' } },
                { "client.initial": { $regex: searchValue, $options: 'i' } },
                { "zone.name": { $regex: searchValue, $options: 'i' } }
            ]
        }

        if (branch) { query['_id'] = branch; }
        if (client) { query['client._id'] = client; }
        if (zone) { query['zone._id'] = zone; }

        const BranchResult = await BranchServices.aggregate([
            { $match: query },
            { $sort: { 'name': 1 } },
            { $skip: req.local.startIndex },
            { $limit: req.local.limit },
            ...(withCount ? [{
                $addFields: {
                    id: { $toString: "$_id" }
                }
            },
            {
                $lookup: {
                    from: "cases",
                    let: { branchId: "$id", },
                    pipeline: [
                        {
                            $match: {
                                $expr: {
                                    $eq: ["$$branchId", "$branch._id"],
                                },
                            },
                        },
                        {
                            $project: { _id: 1, caseNumber: 1, type: 1 }
                        }
                    ],
                    as: "cases",
                }
            },
            {
                $addFields: {
                    cases: {
                        $size: {
                            $filter: {
                                input: "$cases",
                                as: "case",
                                cond: {
                                    $and: [
                                        { $eq: ["$$case.type", "CASE"] }
                                    ],
                                },
                            },
                        },
                    },
                    leads: {
                        $size: {
                            $filter: {
                                input: "$cases",
                                as: "case",
                                cond: {
                                    $and: [
                                        { $ne: ["$$case.type", "CASE"] }
                                    ],
                                },
                            },
                        },
                    },
                }
            }] : []),
        ]);
        if (!BranchResult.status || BranchResult.error) return res.status(500).send(BranchResult);

        const BranchResultCount = await BranchServices.count(query);
        if (!BranchResultCount.status || BranchResultCount.error) return res.status(500).send(BranchResultCount);

        count = BranchResultCount.count

        return res.status(200).send({ status: true, branches: BranchResult.data, count: count });
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.getAllBranches= ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: e.toString() });
    }
};

/**
 * @description This function is used for get list of Branches without Zone. 
 * @param {*} req 
 * @param {*} res 
 * @returns Object BranchResult { status : true, branches : array }
 * - { status: false, error: string }
 */
module.exports.getAllBranchesWithoutZone = async (req, res) => {
    try {

        const clientId = req.params.id
        const orgId = req.local.user.organization._id;
        let query = {
            // 'organization._id': orgId, 
            "client._id": clientId,
            "zone": { "$exists": false }
        };

        const BranchResult = await BranchServices.find({ query, limit: 1000000 });
        if (!BranchResult.status || BranchResult.error) return res.status(500).send(BranchResult);

        return res.status(200).send({ status: BranchResult.status, branches: BranchResult.data });
    }
    catch (e) {
        console.log("Error: ", e);
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};
/**
 * @description This function is used for get one branch from requested branch id.  
 * @param {String} req.params from that get branch id. 
 * @param {*} res 
 * @returns Object BranchResult { status : true, branches : object }
 * - { status: false, error: string }
 */
module.exports.getBranch = async (req, res) => {
    try {
        const id = req.params.id;

        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const BranchResult = await BranchServices.findById(id);
        if (!BranchResult.status || BranchResult.error) return res.status(500).send(BranchResult);

        return res.status(200).send({ status: BranchResult.status, branches: BranchResult.data });
    }
    catch (e) {
        console.log("Error: ", e);
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This function is used for create branch.
 * @param {Object} req.local.branch i.e. requested branch object 
 * @param {*} res 
 * @returns Object branch_res { status : true, branches : object }
 * - { status: false, error: string }
 */
module.exports.createBranch = async (req, res) => {
    try {
        const datas = req.local.branch;

        const alreadyCreated = await BranchServices.find({ query: { "client._id": datas.client, name: datas.name }, limit: 1000000 })
        if (!alreadyCreated.status || alreadyCreated.error) return res.status(500).send(alreadyCreated);

        if (alreadyCreated.count == 0) {

            const organization = req.local.user.organization._id;

            const Organization = await OrganizationServices.getById(organization);
            if (!Organization.status) return res.status(500).send(Organization);

            datas.organization = Organization.Organizations;

            if (datas.pickupPerson) {
                const pickupPerson = await UserServices.findById(datas.pickupPerson);
                if (!pickupPerson.status) return res.status(500).send(pickupPerson);

                datas.pickupPerson = pickupPerson.users;
            }

            if (datas.client) {
                const client = await ClientServices.findById(datas.client);
                if (!client.status) return res.status(500).send(client);

                datas.client = client.data;
            }
            if (datas.zone) {
                const ZoneResult = await ZoneServices.getZone(datas.zone);
                if (!ZoneResult.status || ZoneResult.error) return res.status(500).send(ZoneResult);

                datas.zone = ZoneResult.zones;

            }
            const branch_res = await BranchServices.create(datas);
            if (!branch_res.status) return res.status(500).send(branch_res);

            // // create creditOfficer   

            // if (datas.creditOfficers) {

            //     await Promise.all(datas.creditOfficers.map(async (value) => {
            //         value.branch = branch_res.branches;
            //         value.organization = Organization.Organizations;

            //         const creditOfficer = await CreditOfficerServices.create(value);
            //     }))
            // }

            return res.status(200).send({ status: branch_res.status, branches: branch_res.data });
        }
        else {
            return res.status(500).send({ status: false, error: "Already Created!" });
        }
    }

    catch (err) {
        console.log(err);
        return res.status(500).send({ status: false, error: err });
    }
};
/**
 * @description This function is used for update branch data
 * @param {*} req.params from that get branch id and from req.local.branch get branch object
 * @param {*} res 
 * @returns Object branch_res { status : true, branches : object }
 * - { status: false, error: string }
 */
module.exports.updateBranch = async (req, res) => {
    try {
        const id = req.params.id;

        const datas = req.local.branch;

        const organization = req.local.user.organization._id;

        const Organization = await OrganizationServices.getById(organization);

        if (datas.pickupPerson) {
            const pickupPerson = await UserServices.findById(datas.pickupPerson);
            if (!pickupPerson.status) return res.status(500).send(pickupPerson);

            datas.pickupPerson = pickupPerson.users;
        }
        if (datas.client) {
            const client = await ClientServices.findById(datas.client);
            if (!client.status) return res.status(500).send(client);

            datas.client = client.data;
        }
        if (datas.zone) {
            const ZoneResult = await ZoneServices.getZone(datas.zone);
            if (!ZoneResult.status || ZoneResult.error) return res.status(500).send(ZoneResult);

            datas.zone = ZoneResult.zones;
        }

        const branch_res = await BranchServices.updateById(id, datas);
        if (!branch_res.status) return res.status(500).send(branch_res);

        // create creditOfficer 
        if (datas.creditOfficers) {

            await Promise.all(datas.creditOfficers.map(async (value) => {

                if (!("_id" in value)) {
                    datas.organization = Organization.Organizations;
                    value.branch = branch_res.data;

                    const creditOfficer = await CreditOfficerServices.create(value);
                }
            }))
        }
        return res.status(200).send({ status: branch_res.status, branches: branch_res.data });
    }
    catch (err) {
        console.log(err)
        return res.status(500).send({ status: false, error: err });
    }
};
/**
 * @description This function is used for get all creditOfficer for particular branch 
 * @param {String} req.params from that get branch id 
 * @param {*} res 
 * @returns  Object userResult { status: true, users: array }
 * - { status: false, error: string }
 */
module.exports.getAllCreditOfficersByBranch = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const userResult = await CreditOfficerServices.find({ query: { "branch._id": id, isDeleted: false }, limit: 500 });
        if (!userResult.status || userResult.error) return res.status(500).send(userResult);

        return res.status(200).send({ status: true, users: userResult.data });
    }
    catch (e) {
        console.log("Error: ", e);
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};
/**
 * @description This function is used for get all deactivated creditOfficer branch wise 
 * @param {String} req.params from that get branch id 
 * @param {*} res 
 * @returns Object branchResult {
    "status": true,
    "TransferBranchHistory": arr}
 */
module.exports.getAllDeactivatedCreditOfficersByBranch = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const branchResult = await TransferBranchHistoryServices.getAll({ "oldBranch._id": id });
        if (!branchResult.status || branchResult.error) return res.status(500).send(branchResult);

        return res.status(200).send(branchResult);
    }
    catch (e) {
        console.log("Error: ", e);
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};
module.exports.deleteBranch = async (req, res) => {
    // return res.status(200).send({ message: "Hi There Its  Admin" });
};
/**
 * @description This function is used for put zone into multiple branches.
 * @param {Object} req.body from that { "zone":"zoneId", "branches": ["branchId", "branchId"]}
 * @param {*} res 
 * @returns Object { status: true, message: "COMPLETED!" }
 */
module.exports.branchUtility = async (req, res) => {
    try {

        const datas = req.body;
        var zoneObj;
        if (datas.zone) {
            const ZoneResult = await ZoneServices.getZone(datas.zone);
            if (!ZoneResult.status || ZoneResult.error) return res.status(500).send(ZoneResult);

            zoneObj = ZoneResult.zones;
        }
        await datas.branches.map(async (id) => {

            let BranchResult = await BranchServices.findById(id);

            var branchObj = BranchResult.data.toJSON();
            branchObj.zone = zoneObj;

            const branch_res = await BranchServices.updateById(id, branchObj);

        })
        return res.status(200).send({ status: true, message: "COMPLETED!" });
    }
    catch (err) {
        console.log(err)
        return res.status(500).send({ status: false, error: err });
    }
};