const fs = require("fs");
const path = require("path");
const moment = require('moment');
const { ObjectID } = require("mongodb");
const _ = require('loadsh');

const PaymentServices = require("../Services/Payment.services");
const CaseServices = require("../Services/Case.services");
const nInvoiceServices = require('../Services/CreateServices')('nInvoice');
const UtilitiesController = require("./Utilities.controller");

const CONFIG = require("../Constants/Config");
const Case = require("../Model/Case");
const Payment = require("../Model/Payment");
const nInvoice = require("../Model/nInvoice");
const { logger } = require("../logger");
const { dateRange, generateExcel } = require('../Utils/Utility');

/**
 * @description This is used for get list of all payments
 * @param {*} req
 * @param {*} res
 * @returns Object PaymentResult { status: true, payments: Array }
 * - { status : false, error : string }
 */
module.exports.getDashboardPaymentCount = async (req, res) => {
    try {
        const user = req.local.user;
        let obWorkTypes = [];
        const orgBranchPermission = user.orgBranches || [];
        const { start, end } = req.query;

        let obQuery = orgBranchPermission.map((ob, id) => {
            obWorkTypes = _.union(obWorkTypes, ob.workTypes);
            return { $and: [{ organizationId: ob._id }, { workType: { $in: ob.workTypes } }] };
        })
        openQuery = { status: "OPEN", ...(obQuery.length > 0 ? { $or: obQuery } : {}) }
        // const userId = req.query.userId || null;
        // let query = { 'users.userId': ObjectID(userId ? userId : user._id) };
        let responseCounts = {
            PENDING: { count: 0 },
            IN_PROCESS: { count: 0 },
            REVIEW: { count: 0 },
            COMPLETED: { count: 0 }
        };

        const PaymentCount = await PaymentServices.aggregate([
            {
                $match: {

                    $or: [
                        {
                            $and: [
                                { paymentStatus: "PENDING" }
                            ]
                        },
                        {
                            $and: [
                                { paymentStatus: "IN_PROCESS" },
                                { updatedAt: dateRange(start, end) }
                            ]
                        },
                        {
                            $and: [
                                { paymentStatus: "REVIEW" },
                                { updatedAt: dateRange(start, end) }
                            ]
                        },
                        {
                            $and: [
                                { paymentStatus: "COMPLETED" },
                                { updatedAt: dateRange(start, end) }
                            ]
                        },
                    ],
                },
            },
            { $group: { _id: "$paymentStatus", count: { $sum: 1 } } },
        ]);
        if (!PaymentCount.status || PaymentCount.error) return res.status(500).send(PaymentCount);

        const statusCounts = PaymentCount.payments.map((st, id) => {
            responseCounts[st._id].count = st.count;
        });

        return res.status(200).send({ status: true, counts: responseCounts });
    }
    catch (e) {
        logger.error("🚀 ~ file: Payment.controller.js:48 ~ module.exports.getAllPayments= ~ e:" + e.toString())
        return res.status(500).send({ status: false, error: e.toString() });
    }
};

/**
 * @description This is used for get list of all Payments
 * @param {*} req 
 * @param {*} res 
 * @returns Object PaymentResult { status: true, Payments: Array }
 * - { status : false, error : string }
 */
module.exports.getAllPayments = async (req, res) => {
    try {
        const user = req.local.user;
        let { searchText, start, end, paymentStatus = "PENDING" } = req.query;

        let query = {
            paymentStatus: paymentStatus,
            ...(paymentStatus == "COMPLETED" ? { createdAt: dateRange(start, end) } : {})
        };
        let searchRegex = {};

        if (!!searchText) {
            searchRegex = {
                '$or': [
                    { 'caseIds.caseNumber': { $regex: searchText, $options: 'i' } },
                    { 'caseIds.ownerName': { $regex: searchText, $options: 'i' } },
                    { "caseIds.typeOfReport.label": { $regex: searchText, $options: 'i' } },
                ]
            }
        }

        const PaymentCountResult = await PaymentServices.aggregate([
            { '$match': query },
            { '$lookup': { 'from': 'cases', 'localField': 'caseIds', 'foreignField': '_id', 'as': 'caseIds' } },
            { '$match': searchRegex },
            { '$count': "count" }
        ]);

        if (!PaymentCountResult.status || PaymentCountResult.error) return res.status(500).send(PaymentCountResult);
        const PaymentCount = (PaymentCountResult.payments.length > 0 && PaymentCountResult.payments[0].count) ? PaymentCountResult.payments[0].count : 0

        const PaymentResult = await PaymentServices.aggregate(
            [
                { '$match': query },
                { '$lookup': { 'from': 'cases', 'localField': 'caseIds', 'foreignField': '_id', 'as': 'caseIds' } },
                // { '$unwind': { 'path': '$caseIds' } },
                // { '$lookup': { 'from': 'organizations', 'localField': 'organizationId', 'foreignField': '_id', 'as': 'organizationId' } },
                { '$match': searchRegex },
                { '$sort': { "createdAt": -1 }, },
                { "$skip": req.local.startIndex },
                { "$limit": req.local.limit },
            ]
        );

        if (!PaymentResult.status || PaymentResult.error) return res.status(500).send(PaymentResult);
        return res.status(200).send({ status: true, Payments: PaymentResult.payments || [], count: PaymentCount });
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.getAllPayments= ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: e.toString() });
    }
};

/**
 * @description This is used for get one payment
 * @param {*} req.params.id from that get paymentId
 * @param {*} res
 * @returns Object PaymentResult { status: true, PaymentResult, Empanelments: Object.values(arr) }
 * - { status : false, error : string }
 */
module.exports.getPayment = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const PaymentResult = await PaymentServices.aggregate(
            [
                {
                    $match: {
                        _id: ObjectID(id),
                    },
                },
                // {
                //     $unwind: {
                //         path: "$caseIds",
                //     },
                // },
                {
                    $lookup: {
                        from: "cases",
                        localField: "caseIds",
                        foreignField: "_id",
                        as: "caseIds",
                    },
                },
                {
                    $lookup: {
                        from: "ninvoices",
                        localField: "invoiceIds",
                        foreignField: "_id",
                        as: "invoiceIds",
                    },
                },
                {
                    $lookup: {
                        from: "ninvoices",
                        localField: "invoiceIds.taxInvoiceId",
                        foreignField: "_id",
                        as: "taxInvoiceIds",
                    },
                },
            ]
        );

        if (!PaymentResult.status || PaymentResult.error) return res.status(500).send(PaymentResult);

        if (PaymentResult.payments.length == 0) return res.status(404).send({ status: true, payments: {} })

        return res.status(200).send({ status: true, payments: PaymentResult.payments[0] });
    } catch (e) {
        logger.error(`🚀 ~ module.exports.getPayment= ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};


module.exports.pendingInvoiceCaseList = async (req, res) => {
    const user = req.local.user;
    const { clientId, branchId, searchText, cases, clientType } = req.query;
    const ary = (cases && cases.split(',')) && cases.split(',')?.map((id, index) => id?.trim()).filter((id, index) => !!id) || [];
    const selectedCases = ary.map((cid, id) => ObjectID(cid))

    let query = {
        ...(!!req.query.organization && { 'organization._id': req.query.organization }),
        // 'type': 'CASE',
        'caseStatus': { $nin: ["CANCELLED", "COMPLETED"] },
        'paymentStatus': { $nin: ["COMPLETED"] },
        // invoiceId: { $exists: true }
    }

    if (clientType !== "PRIVATE_CLIENT" && clientId) { query = { ...query, 'client._id': clientId } }
    // if (clientType !== "PRIVATE_CLIENT" && branchId) { query = { ...query, 'branch._id': branchId } }
    let searchRegex = {};

    if (!!searchText) {
        searchRegex = {
            '$or': [
                { 'caseNumber': { $regex: searchText, $options: 'i' } },
                { 'ownerName': { $regex: searchText, $options: 'i' } },
                { 'proposePurchaserName': { $regex: searchText, $options: 'i' } },
                { "typeOfReport.label": { $regex: searchText, $options: 'i' } },
            ]
        }
    }
    try {
        // for already payment request raised so can remove that from list
        const processInvoiceCases = await Payment.aggregate([
            {
                $match: {
                    paymentStatus: {
                        $in: ["IN_PROCESS", "PENDING"],
                    },
                },
            },
            {
                $group: {
                    _id: null,
                    caseIds: {
                        $push: "$caseIds",
                    },
                },
            },
            {
                $addFields: {
                    caseIds: {
                        $reduce: {
                            input: "$caseIds",
                            initialValue: [],
                            in: {
                                $concatArrays: ["$$value", "$$this"],
                            },
                        },
                    },
                }
            },
            {
                $project: { caseIds: 1 }
            }
        ]).allowDiskUse(true);

        const processInvoices = (processInvoiceCases.length > 0 && processInvoiceCases[0]?.caseIds) ? processInvoiceCases[0]?.caseIds : []

        const casesResult = await Case.aggregate([
            {
                $match: {
                    _id: { $nin: [...processInvoices, ...(selectedCases.length > 0 ? selectedCases : [])] },
                    ...query,
                    ...searchRegex
                }
            },
            {
                $sort: { createdAt: -1 }
            },
            {
                $addFields: {
                    caseValue: {
                        $cond: {
                            if: { $ne: ["$valuation", null] },
                            then: "$valuation",
                            else: {
                                $cond: {
                                    if: { $ne: ["$valuation.finalValue", null] },
                                    then: "$valuation.finalValue",
                                    else: {
                                        $cond: {
                                            if: {
                                                $and: [
                                                    { $ne: ["$externalReportDetails.rate", null] },
                                                    { $ne: ["$externalReportDetails.area", null] }
                                                ]
                                            },
                                            then: {
                                                $multiply: [
                                                    { $toDouble: "$externalReportDetails.rate" },
                                                    { $toDouble: "$externalReportDetails.area" }
                                                ]
                                            },
                                            else: "0.00"
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            },
            {
                $project: {
                    'ownerName': 1, 'typeOfReport': 1, 'clientType': 1, 'client': 1, 'branch': 1, 'organization': 1, 'propertyType': 1,
                    'caseNumber': 1, 'proposePurchaserName': 1, 'proposePurchasers': 1, externalReportDetails: 1, flatRate: 1, 'valuation.finalValue': 1, 'valuation.finaltotal': 1,
                    'caseValue': 1
                }
            },
            { "$skip": req.local.startIndex },
            { "$limit": req.local.limit },
        ]).allowDiskUse(true);

        return res.status(200).send({ status: true, cases: casesResult });
    } catch (e) {
        logger.error(`🚀 ~ file: Payment.controller.js:302 ~ module.exports.pendingInvoiceCaseList= ~ e: ${e.toString()}`);
        return res.status(500).send({ status: false, error: e.toString() });
    }
};

module.exports.pendingPaymentCaseList = async (req, res) => {
    const user = req.local.user;
    const { clientId, branchId, searchText } = req.query;

    let query = {
        ...(!!req.query.organization && { 'organization._id': req.query.organization }),
        'type': 'CASE',
        'caseStatus': { $nin: ["CANCELLED", "COMPLETED"] },
        // invoiceId: { $exists: true }
    }

    if (clientId) { query = { ...query, client: clientId } }
    if (branchId) { query = { ...query, branch: branchId } }
    let searchRegex = {};

    if (!!searchText) {
        searchRegex = {
            '$or': [
                { 'invoiceNo': { $regex: searchText, $options: 'i' } },
                { 'cases.caseNumber': { $regex: searchText, $options: 'i' } },
                { 'cases.ownerName': { $regex: searchText, $options: 'i' } },
                { 'cases.outward.number': { $regex: searchText, $options: 'i' } },
                { "cases.typeOfReport.label": { $regex: searchText, $options: 'i' } },
            ]
        }
    }
    try {
        const pendingPaymentInvoices = await nInvoice.aggregate([
            {
                $match: {
                    ...searchRegex,
                    invoiceStatus: "COMPLETED",
                    paymentStatus: { $in: ["PENDING", "PARTIAL_RECEIVED"] },
                },
            },
            {
                $sort: { createdAt: -1 }
            },
            { "$skip": req.local.startIndex },
            { "$limit": req.local.limit },
        ]).allowDiskUse(true);

        return res.send({ status: true, invoices: pendingPaymentInvoices });
    } catch (e) {
        logger.error(`🚀 ~ file: Payment.controller.js:352 ~ module.exports.pendingPaymentCaseList= ~ e: ${e.toString()}`);
        return res.send({ status: false, error: e.toString() });
    }
};

module.exports.outstandingPaymentList = async (req, res) => {
    const user = req.local.user;
    const { clientId, branchId, searchText, download = "false" } = req.query;
    let userQuery = {};

    // if (!['VALUER', 'ADMIN', 'FRONTDESK', 'PRINTING', 'ACCOUNTANT'].includes(user.role)) {
    userQuery['$or'] = [
        { "caseDetails.staff.createdBy": user._id },
        { "caseDetails.staff.pickupPerson._id": user._id },
        { "caseDetails.staff.maker._id": user._id },
        { "caseDetails.staff.checker._id": user._id },
        { "caseDetails.staff.valuers._id": user._id },
        { "caseDetails.staff.rateVerifier._id": user._id },
        { "caseDetails.staff.siteEngineer._id": user._id }
    ]
    // }

    let query = {
        ...(!!req.query.organization && { 'organization._id': req.query.organization }),
    }
    let searchRegex = { '$or': [] };
    let groupQuery = {
        _id: { clientId: "$client._id" },
        client: { $first: "$client" },
        invoices: { $addToSet: "$$ROOT" },
        pendingAmount: { $sum: "$pendingAmount" },
    };

    if (clientId) {
        query = { ...query, 'client._id': ObjectID(clientId) }
        groupQuery._id = { clientId: "$client._id", branchId: "$branch._id" };
        groupQuery.branch = { $first: '$branch' };

        // search text apply for branch name and branch address
        if (!!searchText && !branchId) {
            searchRegex['$or'] = [
                { 'branch.name': { $regex: searchText, $options: 'i' } }
            ];
        }
    }
    else {
        if (!!searchText) {
            searchRegex['$or'] = [
                { 'client.name': { $regex: searchText, $options: 'i' } },
                { "client.initial": { $regex: searchText, $options: 'i' } },
            ];
        }
    }


    if (branchId) {
        query = { ...query, 'branch._id': ObjectID(branchId) }

        // search text apply only on invoice and case.
        searchRegex['$or'] = [
            { 'invoiceNo': { $regex: searchText, $options: 'i' } },
            { 'cases.caseNumber': { $regex: searchText, $options: 'i' } },
            { 'cases.ownerName': { $regex: searchText, $options: 'i' } },
            { 'cases.outward.number': { $regex: searchText, $options: 'i' } },
            { "cases.typeOfReport.label": { $regex: searchText, $options: 'i' } },
        ]
    }
    console.log("query:", JSON.stringify(searchRegex));
    try {
        let aggregation = [
            {
                $unwind: { path: "$cases" },
            },
            {
                $match: {
                    paymentStatus: { $ne: "RECEIVED" },
                    invoiceStatus: "COMPLETED",
                    ...query,
                    ...((searchRegex['$or'].length > 0) ? searchRegex : []),
                },
            }
        ];

        if (!!download && download == "true") {
            aggregation = [
                ...aggregation,
                {
                    $lookup: {
                        from: "cases",
                        localField: "cases._id",
                        foreignField: "_id",
                        as: "caseDetails",
                    },
                },
                { $unwind: { path: "$caseDetails" } },
                { $match: { '$and': [userQuery] } },
                {
                    $project: {
                        client: "$client.name",
                        Bank: "$client.name",
                        Branch: "$branch.name",
                        "Owner Name": "$caseDetails.ownerName",
                        Address: "$caseDetails.fullAddress",
                        caseNumber: "$caseDetails.caseNumber",
                        "Case Referred By": "$caseDetails.refereedBy",
                        "Field Staff": {
                            $concat: [
                                "$caseDetails.staff.pickupPerson.firstName",
                                " ",
                                "$caseDetails.staff.pickupPerson.lastName"
                            ],
                        },
                        "Outward Date": "$caseDetails.outward.date",
                        "outward No": "$caseDetails.outward.number",
                        invoiceNo: "$invoiceNo",
                        amount: "$cases.fees",
                        invoiceAmount: "$pendingAmount",
                        status: "PENDING",
                    }
                }
            ]
            const pendingPaymentInvoices = await nInvoice.aggregate(aggregation).allowDiskUse(true);
            const filePath = await generateExcel(pendingPaymentInvoices, 'outstanding-payment.xlsx');
            logger.error(`🚀 ~ module.exports.outstandingPaymentList= ~ filePath:` + filePath)

            return res.sendFile(filePath, {}, (err) => {
                if (err) {
                    logger.error('Error occurred while sending file:' + err.toString());
                    return res.status(err.status).end();
                } else {
                    // fs.unlinkSync(filePath); // Remove the temporary file after sending
                }
            });
        } else {
            aggregation = [
                ...aggregation,
                { $group: groupQuery },
                {
                    $sort: {
                        "branch.name": 1,
                        'client.name': 1,
                    }
                },
                { "$skip": req.local.startIndex },
                { "$limit": req.local.limit }
            ];
            const pendingPaymentInvoices = await nInvoice.aggregate(aggregation).allowDiskUse(true);
            return res.send({ status: true, invoices: pendingPaymentInvoices });
        }

    } catch (e) {
        logger.error(`🚀 ~ file: Payment.controller.js:352 ~ module.exports.pendingPaymentCaseList= ~ e: ${e.toString()}`);
        return res.send({ status: false, error: e.toString() });
    }
};

module.exports.addPayment = async (req, res) => {
    try {
        const paymentBody = req.local.payment;
        let currentPath, fileName;

        let paymentData = {
            ...paymentBody,
            paymentType: paymentBody.paymentType.toUpperCase(),
            createdBy: req.local.user
        };
        if (!paymentBody.number) {
            paymentBody.number = moment(new Date()).date().toString() + moment(new Date()).month().toString() + "_" + paymentBody.number;
        }

        // if has file then copy the payment in the case payment folder
        if (req.file) {
            currentPath = process.cwd() + "/" + req.file.path;
            fileName = "payment_" + req.file.filename;

            if (!fs.existsSync(process.env.CDN_PATH + "/payments/")) {
                fs.mkdirSync(process.env.CDN_PATH + "/payments/", { recursive: true });
            }
            const destination = path.join(process.env.CDN_PATH, "payments", fileName);

            await UtilitiesController.copyFile(currentPath, destination);
            const filepath = path.join("payments", fileName);

            paymentData = {
                ...paymentData,
                paymentFile: {
                    ...req.file,
                    fileName,
                    path: filepath,
                    documentType: { key: "PAYMENT", value: "PAYMENT" },
                    uploadedBy: req.local.user,
                    uploadedAt: new Date(),
                },
            };
        }

        const PaymentResult = await PaymentServices.createPayment(paymentData);
        // const paymentCollectPaymentRes = await Payment.create({ workType: workType.PAYMENT_COLLECT, caseIdsString, organizationId: cs.organization._id });

        return res.status(200).send(PaymentResult);
    } catch (e) {
        logger.error(`🚀 ~ module.exports.addPayment= ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: e.toString() });
    }
};

module.exports.updatePayment = async (req, res) => {
    try {
        const user = req.local.user;
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        let payment = req.local.payment
        // paymentSettlement, voucher

        // if has file then copy the payment in the case payment folder
        if (req.file) {
            currentPath = process.cwd() + "/" + req.file.path;
            fileName = "payment_" + req.file.filename;

            if (!fs.existsSync(process.env.CDN_PATH + "/payments/")) {
                fs.mkdirSync(process.env.CDN_PATH + "/payments/", { recursive: true });
            }
            const destination = path.join(process.env.CDN_PATH, "payments", fileName);

            await UtilitiesController.copyFile(currentPath, destination);
            const filepath = path.join("payments", fileName);

            payment.paymentFile = {
                ...req.file,
                fileName,
                path: filepath,
                documentType: { key: "PAYMENT", value: "PAYMENT" },
                uploadedBy: req.local.user,
                uploadedAt: new Date(),
            };
        }

        if (!!payment.remark) {
            payment = {
                ...payment, $push: {
                    remarks: {
                        remark: payment.remark,
                        createdBy: user,
                        createdAt: new Date()
                    }
                }
            };
        }

        const PaymentUpdateRes = await PaymentServices.updatePayment(id, { ...payment, updatedAt: new Date(), updatedBy: user });
        if (!PaymentUpdateRes.status) return res.status(500).send(PaymentUpdateRes);

        const PaymentRes = await PaymentServices.getPayment(id, { populate: [{ path: 'invoiceIds' }, { path: 'caseIds' }] });
        if (!PaymentRes.status || PaymentRes.error) return res.status(500).send(PaymentRes);
        const paymentResult = PaymentRes.payments.toJSON();

        return res.status(200).send({ status: true, payments: paymentResult });
    } catch (err) {
        logger.error("🚀 ~ file: Payment.controller.js:436 ~ module.exports.paymentSettlement= ~ err:" + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
}

module.exports.updatePaymentStatus = async (req, res) => {
    try {
        const user = req.local.user;
        const { id, status } = req.params;

        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Payment Not Found" });
        if (!(status && status.length != 0)) return res.status(404).send({ status: false, error: "Unknown Status" });

        let remark = {
            remark: "",
            createdBy: user,
            createdAt: new Date()
        };

        switch (status.toLowerCase()) {
            case "collected":
                remark.remark = `Cash is ${status} by ${(user.firstName || "") + " " + (user.lastName || "")}`;
                break;
            case "deposited":
                remark.remark = `Cheque is ${status} by ${(user.firstName || "") + " " + (user.lastName || "")}`;
                break;
            case "verified":
                remark.remark = `Online Payment is ${status} by ${(user.firstName || "") + " " + (user.lastName || "")}`;
                break;
            case 'completed':
                remark.remark = `Payment Process is ${status} by ${(user.firstName || "") + " " + (user.lastName || "")}`;
                break;
            default:
                break;
        }

        let updateData = {
            updatedAt: new Date(),
            updatedBy: user,
            collectedAt: new Date(),
            collectedBy: user,
            paymentStatus: "IN_PROCESS",
            $push: { remarks: remark },
        };

        const Payment_res = await PaymentServices.updatePayment(id, updateData);
        if (!Payment_res.status) return res.status(500).send(Payment_res);

        const PaymentResult = await PaymentServices.getPayment(id, { populate: [{ path: 'invoiceIds' }, { path: 'caseIds' }] });
        if (!PaymentResult.status || PaymentResult.error) return res.status(500).send(PaymentResult);
        const paymentRes = PaymentResult.payments.toJSON();

        return res.status(200).send({ status: true, payments: paymentRes });
    } catch (err) {
        logger.error(`🚀 ~ file: Payment.controller.js:502 ~ module.exports.updatePaymentStatus= ~ err: ${err.toString()}`)
        return res.status(500).send({ status: false, error: err.toString() });
    }
}

module.exports.deletePayment = async (req, res) => {
    try {
        const user = req.local.user;
        const { id } = req.params;

        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Payment Not Found" });

        const paymentRes = await PaymentServices.deletePayment({ _id: id, paymentStatus: { $ne: "COMPLETED" } });
        if (!paymentRes.status) return res.status(500).send(paymentRes);

        return res.status(200).send({ status: true, payments: paymentRes });
    } catch (err) {
        logger.error(`🚀 ~ file: Payment.controller.js:502 ~ module.exports.updatePaymentStatus= ~ err: ${err.toString()}`)
        return res.status(500).send({ status: false, error: err.toString() });
    }
}

module.exports.addRemark = async (req, res) => {
    try {
        const user = req.local.user;
        const id = req.params.id;
        const body = req.body;

        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const data = { $push: { 'remarks': { ...body, createdBy: user, createdAt: new Date(), } } };
        data.updatedBy = user._id;

        const paymentRes = await PaymentServices.updatePayment(id, data);
        if (!paymentRes.status) return res.status(500).send(paymentRes);

        return res.status(200).send(paymentRes);
    } catch (err) {
        logger.error(`🚀 ~ file: Payment.controller.js:523 ~ module.exports.addRemark= ~ err: ${err.toString()}`)
        return res.status(500).send({ status: false, error: err.toString() });
    }
};

/**
 * @description This is used to upload tax invoice
 */
module.exports.taxInvoiceUpload = async (req, res) => {
    try {
        const id = req.params.id; // payment id
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });
        if (!req.file) { return res.status(500).send({ status: false, error: "No Invoice File Found" }); }
        const { invoiceId, invoicedAmount, invoiceNo, voucher } = req.local.invoice;

        // get payment
        const PaymentResult = await PaymentServices.getPayment(id);
        if (!PaymentResult.status || PaymentResult.error) return res.status(500).send(PaymentResult);
        const paymentRes = PaymentResult.payments.toJSON();

        // get proforma invoice
        const InvoiceResult = await nInvoiceServices.findById(invoiceId);
        if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);

        // make new invoice for tax invoice
        let newInv = InvoiceResult.data.toJSON();
        delete newInv._id;
        delete newInv.createdAt;
        delete newInv.updateAt;
        delete newInv.updatedBy;
        delete newInv.invoiceFile;

        const fileName = "inv_" + req.file.filename;
        let currentPath = process.cwd() + "/" + req.file.path;

        if (!fs.existsSync(process.env.CDN_PATH + '/invoices/' + id)) {
            fs.mkdirSync(process.env.CDN_PATH + '/invoices/' + id, { recursive: true });
        }
        const destination = path.join(process.env.CDN_PATH, "invoices", id, fileName);
        await UtilitiesController.copyFile(currentPath, destination);
        const filepath = path.join("invoices", id, fileName);

        newInv = {
            ...newInv,
            status: true,
            invoiceStatus: "COMPLETED",
            amount: invoicedAmount,
            invoiceNo: invoiceNo,
            completedAt: new Date(),
            invoiceType: "TAX",
            invoiceFile: {
                ...req.file,
                fileName,
                path: filepath,
                documentType: { key: "TAX_INVOICE", value: "TAX_INVOICE" },
                uploadedBy: req.local.user,
                uploadedAt: new Date()
            },
            ...(voucher && { voucher: voucher }),
            payments: [
                { _id: paymentRes._id, receivedAmount: paymentRes.amount, paymentSettlement: paymentRes.paymentSettlement }
            ],
            paymentStatus: "RECEIVED"
        };

        // create new tax invoice
        const newInvResult = await nInvoiceServices.create({ ...newInv });
        if (!newInvResult.status) return res.status(500).send(newInvResult);
        const newInvoice = newInvResult.data;

        // update proforma invoice with tax invoice id
        const updatePROFORMAInvoiceResult = await nInvoiceServices.updateById(invoiceId, { taxInvoiceId: newInvoice._id });

        return res.status(200).send({ status: true, payments: paymentRes });
    }
    catch (err) {
        logger.error(`🚀 ~ file: Payment.controller.js:659 ~ module.exports.taxInvoiceUpload= ~ err: ${err.toString()}`)
        return res.status(500).send({ status: false, error: err.toString() });
    }
};

/**
 * @description This is used to complete payment
 */
module.exports.completePayment = async (req, res) => {
    try {
        const id = req.params.id; // payment id
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });
        const user = req.local.user;

        const PaymentResult = await PaymentServices.getPayment(id);
        if (!PaymentResult.status || PaymentResult.error) return res.status(500).send(PaymentResult);
        const paymentRes = PaymentResult.payments.toJSON();

        const invoices = paymentRes.paymentSettlement;
        await Promise.all(invoices.map(async (inv) => {

            const invoiceResponse = await nInvoiceServices.findById(inv._id);
            const invDetails = invoiceResponse.data;
            if (!invDetails) return;

            //1. update cases and count invoice totals
            await Promise.all(inv.cases.map(async (_case, cid) => {

                const caseID = _case._id.toString();
                let caseIndex = invDetails.cases.findIndex(cs => cs._id.toString() === caseID);
                // console.log("=> ", caseIndex, caseID, invDetails.cases[caseIndex])
                if (caseIndex == -1) return;
                invDetails.cases[caseIndex].invoicedAmount = _case.fees;
                invDetails.cases[caseIndex].isCompleted = _case.isCompleted;
                invDetails.cases[caseIndex].receivedAmount += +_case.receivedAmount;
                invDetails.cases[caseIndex].pendingAmount = +_case.pendingAmount;
                // console.log("=> updated value", invDetails.cases[caseIndex])

                let newCaseData = { $push: { paymentIds: paymentRes._id } }
                // update case paymentStatus according to invoice type => if ADV_ then don't COMPLETE else COMPLETE
                if (_case.isCompleted && (inv.invoiceType === "PROFORMA" || inv.invoiceType === "TAX")) {
                    newCaseData = { ...newCaseData, paymentStatus: "RECEIVED" }
                } else {
                    newCaseData = { ...newCaseData, paymentStatus: "PARTIAL_RECEIVED" }
                }

                await CaseServices.updateCase(caseID, newCaseData);

                // if case isComplete then add to completedCaseIds of Invoice object
                if (!!inv.isCompleted) {
                    invDetails.completedCaseIds.push(inv._id);
                }
            }));

            // 2. add payment details to invoice
            invDetails.payments.push({ _id: paymentRes._id, receivedAmount: paymentRes.paymentTotalAmount, paymentSettlement: invoices });

            // 3. update receivedAmount, pendingAmount, 
            invDetails.receivedAmount += +inv.receivedAmount;
            invDetails.pendingAmount = invDetails.pendingAmount;
            invDetails.accounting.push({ paymentId: paymentRes._id, ...inv, completedAt: new Date(), completedBy: user });

            if (inv.isCompleted || invDetails.pendingAmount == 0) {
                invDetails.paymentStatus = "RECEIVED";
            } else {
                invDetails.paymentStatus = "PARTIAL_RECEIVED";
            }

            logger.error(`🚀 ~ awaitPromise.all ~ invDetails:` + JSON.stringify(invDetails))
            await invDetails.save();
        }));

        // const PaymentUpdateRes = await PaymentServices.updatePayment(id, { });
        const PaymentUpdateRes = await PaymentServices.updatePayment(id, { paymentStatus: "COMPLETED", completedBy: user, completedAt: new Date() });
        if (!PaymentUpdateRes.status) return res.status(500).send(PaymentUpdateRes);

        return res.status(200).send({ status: true, payments: PaymentUpdateRes.payments.toJSON() });
    }
    catch (err) {
        logger.error(`🚀 ~ file: Payment.controller.js:659 ~ module.exports.taxInvoiceUpload= ~ err: ${err.toString()}`)
        return res.status(500).send({ status: false, error: err.toString() });
    }
};

// tds total count and amount
module.exports.tdsCount = async (req, res) => {
    try {

        const accountingResult = await PaymentServices.aggregate([
            {
                $match: {
                    paymentStatus: "COMPLETED",
                    'tds.amount': { $exists: true },
                    'tdsStatus': { $in: ["PENDING", "RECEIVED"] },
                }
            },
            {
                $group: {
                    _id: "$tdsStatus",
                    totalAmount: { $sum: "$tds.amount" },
                    count: { $sum: 1 },
                },
            },
            {
                $addFields: { tdsStatus: "$_id" }
            },
        ], { allowDiskUse: true });

        if (!accountingResult.status) return res.status(500).send(accountingResult);
        return res.status(200).send({ status: true, counts: accountingResult.payments || [] });

    } catch (e) {
        logger.error(`🚀 ~ module.exports.tdsCount= ~ e:`, e.toString())
        return res.status(500).send({ status: false, error: e.toString() });
    }
}

module.exports.tdsList = async (req, res) => {
    try {
        const { type, tdsStatus, searchText } = req.query;
        const { startIndex, limit } = req.local;
        let query = {
            paymentStatus: "COMPLETED",
            'tds.amount': { $exists: true },
            ...(!!tdsStatus && { tdsStatus: tdsStatus })
        };
        let searchQuery = {}

        if (searchText) {
            searchQuery['$and'] = [{
                '$or': [
                    { "voucher.narration": { $regex: searchText, $options: 'i' } },
                    { "voucher.party_name": { $regex: searchText, $options: 'i' } },
                    { "voucher.againstRef.name": { $regex: searchText, $options: 'i' } },
                ]
            }];
        }

        let countResult = await PaymentServices.count(query);

        const accountResult = await PaymentServices.aggregate([
            {
                $match: { ...query, ...searchQuery }
            },
            { $sort: { "createdAt": -1 } },
            { $skip: startIndex },
            { $limit: limit }
        ], { allowDiskUse: true });

        if (!accountResult.status) return res.status(500).send(accountResult);
        return res.status(200).send({ status: true, data: accountResult.payments || [], count: countResult.count || 0 });

    } catch (e) {
        logger.error(`🚀 ~ module.exports.tdsList= ~ e:`, e.toString())
        return res.status(500).send({ status: false, error: e.toString() });
    }
};

/**
 * @description This is used to reverse payment
 */
module.exports.reversePayment = async (req, res) => {
    try {
        const id = req.params.id; // payment id
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });
        const user = {}//req.local.user || {};

        const PaymentResult = await PaymentServices.getPayment(id);
        if (!PaymentResult.status || PaymentResult.error) return res.status(500).send(PaymentResult);
        const paymentRes = PaymentResult.payments.toJSON();

        const invoices = paymentRes.paymentSettlement;
        await Promise.all(invoices.map(async (inv) => {

            const invoiceResponse = await nInvoiceServices.findById(inv._id);
            const invDetails = invoiceResponse.data;
            if (!invDetails) return;

            // 1. Reverse cases and count invoice totals
            await Promise.all(inv.cases.map(async (_case, cid) => {

                const caseID = _case._id.toString();
                let caseIndex = invDetails.cases.findIndex(cs => cs._id.toString() === caseID);
                if (caseIndex == -1) return;

                invDetails.cases[caseIndex].invoicedAmount -= +_case.fees || 0;
                invDetails.cases[caseIndex].isCompleted = false;
                invDetails.cases[caseIndex].receivedAmount -= (+_case.receivedAmount);
                invDetails.cases[caseIndex].pendingAmount += (+_case.receivedAmount);
                logger.error(`🚀 ~ awaitPromise.all ~ invDetails:`, invDetails.cases)

                let newCaseData = { $pull: { paymentIds: paymentRes._id } }
                newCaseData = { ...newCaseData, paymentStatus: "PENDING" }

                await CaseServices.updateCase(caseID, newCaseData);

                const completedIndex = invDetails.completedCaseIds.indexOf(inv._id);
                if (completedIndex !== -1) {
                    invDetails.completedCaseIds.splice(completedIndex, 1);
                }
            }));

            // 2. Remove payment details from invoice
            invDetails.payments = invDetails.payments.filter(payment => payment._id !== paymentRes._id);

            // 3. Update receivedAmount, pendingAmount, 
            invDetails.receivedAmount -= +inv.receivedAmount;
            invDetails.pendingAmount += +inv.receivedAmount;

            invDetails.accounting = invDetails.accounting.filter(account => account.paymentId !== paymentRes._id);

            if (invDetails.pendingAmount > 0) {
                invDetails.paymentStatus = "PENDING";
            }
            console.log(invDetails);
            await invDetails.save();
        }));

        // Update payment status to REVERSED
        const PaymentUpdateRes = await PaymentServices.deletePayment({ _id: id });
        if (!PaymentUpdateRes.status) return res.status(500).send(PaymentUpdateRes);

        return res.status(200).send({ status: true, payments: PaymentUpdateRes.payments });
    }
    catch (err) {
        logger.error(`🚀 ~ file: Payment.controller.js:659 ~ module.exports.reversePayment= ~ err: ${err.toString()}`)
        return res.status(500).send({ status: false, error: err.toString() });
    }
};


// Recursive function to reverse payments and cases linked to invoices
const reverseInvoicePaymentsList = async (invoice, processedInvoices = new Set(), processedPayments = new Set()) => {
    let result = {
        payments: [],
        invoices: [],
        cases: [],
        caseNumbers: [],
    };

    // Exit if invoice or payments are empty or already processed
    if (!invoice || invoice.payments.length === 0 || processedInvoices.has(invoice._id.toString())) return result;

    // Mark this invoice as processed
    processedInvoices.add(invoice._id.toString());
    result.invoices.push(invoice._id.toString());

    // Reset cases related to this invoice
    await Promise.all(invoice.cases.map(async (_case) => {
        const caseID = _case._id.toString();
        let caseIndex = invoice.cases.findIndex(cs => cs._id.toString() === caseID);
        if (caseIndex == -1) return;

        result.cases.push(caseID);
        result.caseNumbers.push(invoice.cases[caseIndex].caseNumber);
    }));
    console.log("reverseinv, res: ", result, invoice.payments.length)

    // Loop through accounting to reverse payments
    for (const account of invoice.payments) {
        // Skip payments that are already processed
        if (processedPayments.has(account._id.toString())) continue;

        // Mark this payment as processed
        processedPayments.add(account._id.toString());

        //if (invoice.accounting.length == 1) continue;
        const relatedInvoicePaymentRes = await PaymentServices.getPayment(account._id.toString());
        const relatedInvoicePaymentInvoice = relatedInvoicePaymentRes.payments;

        if (relatedInvoicePaymentInvoice) {
            console.log("in loop: ", account._id.toString(), !result.payments.includes(account._id.toString()), relatedInvoicePaymentInvoice && !result.payments.includes(account._id.toString()));
            result.payments.push(account._id.toString());

            await Promise.all(relatedInvoicePaymentInvoice.invoiceIds.map(async (inv) => {
                // If the invoice was already processed, skip it
                if (processedInvoices.has(inv.toString())) return;

                console.log("invoice id: ", result.invoices, inv.toString(), processedInvoices.has(inv.toString()));

                const invoiceResponse = await nInvoiceServices.findById(inv.toString());
                const invDetails = invoiceResponse.data;
                if (!invDetails) return {};

                const innerResult = await reverseInvoicePaymentsList(invDetails, processedInvoices, processedPayments);
                result.payments = [...new Set([...result.payments, ...innerResult.payments])];
                result.invoices = [...new Set([...result.invoices, ...innerResult.invoices])];
                result.cases = [...new Set([...result.cases, ...innerResult.cases])];
                result.caseNumbers = [...new Set([...result.caseNumbers, ...innerResult.caseNumbers])];

                return invDetails;

            })).then((r) => r.filter(Boolean));
        }
    }
    console.log("=>", result)
    return result;
};

/**
 * @description This is used to reverse payment
 */
module.exports.resetCaseInvoicePayment = async (req, res) => {
    try {
        const id = req.params.id; // payment id
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });
        const user = {}//req.local.user || {};
        const { soft = "true" } = req.query;

        // get payment details.
        const PaymentResult = await PaymentServices.getPayment(id);
        if (!PaymentResult.status || PaymentResult.error) return res.status(500).send(PaymentResult);
        if (!PaymentResult.payments) return res.status(500).send({ status: false, error: "No Payment Found!" });
        const paymentRes = PaymentResult.payments.toJSON();

        const invoiceIds = paymentRes.paymentSettlement.map((inv, inv_id) => { return inv._id; });

        //get every invoice related to this payment
        let invoices = await Promise.all(invoiceIds.map(async (inv) => {
            const invoiceResponse = await nInvoiceServices.findById(inv);
            const invDetails = invoiceResponse.data;
            if (!invDetails) return {};
            return invDetails;
        })).then((r) => r.filter(Boolean))

        // get every payment related to every invoice
        // const payments = invoices.reduce((prev, invs, invid) => { return [...prev, ...invs.accounting] }, []).map((acc, aid) => acc._id);

        // let cases = invoices.reduce((prev, invs, invid) => { return [...prev, ...invs.cases] }, []);
        // const caseIds = cases.map((cs, cs_id) => cs._id.toString());
        // const caseNumbers = cases.map((cs, cs_id) => cs.caseNumber);

        // get ll paymentids, invoices, cases recursively
        let result = await Promise.all(invoices.map(async (invDetails) => {
            // TODO 1. Reverse cases and count invoice totals (handled in recursive function now)
            return await reverseInvoicePaymentsList(invDetails); // Reverse both payments and cases
        })).then((r) => r);

        result = result.reduce((prev, curr, id) => {
            return {
                payments: [...new Set([...prev.payments, ...curr.payments])],
                invoices: [...new Set([...prev.invoices, ...curr.invoices])],
                cases: [...new Set([...prev.cases, ...curr.cases])],
                caseNumbers: [...new Set([...prev.caseNumbers, ...curr.caseNumbers])]
            }
        }, {
            payments: [],
            invoices: [],
            cases: [],
            caseNumbers: []
        });

        console.log("result: ", result);
        if (soft == "false") {
            // 1. delete result.payments
            await Promise.all(result.payments.map(async (pment) => {
                const PaymentUpdateRes = await PaymentServices.deletePayment({ _id: pment });
                if (!PaymentUpdateRes.status) return res.status(500).send(PaymentUpdateRes);
                console.log("Payment Deleted: ", pment)
                return PaymentUpdateRes;
            }));

            // 2. make pending result.payments
            await Promise.all(result.invoices.map(async (inv) => {
                const invoiceResponse = await nInvoiceServices.findById(inv);
                const invDetails = invoiceResponse.data;
                if (!invDetails) return {};

                //delete taxInvoice if taxInvoiceId available
                if (!!invDetails.taxInvoiceId) {
                    const taxInvoiceDelete = await nInvoiceServices.deleteById(invDetails.taxInvoiceId);
                    if (!taxInvoiceDelete.status) return res.status(500).send(taxInvoiceDelete);
                    console.log("taxInvoice Deleted: ", invDetails.taxInvoiceId)
                }

                await Promise.all(invDetails.cases.map(async (_case, _cid) => {
                    invDetails.cases[_cid].invoicedAmount = invDetails.cases[_cid].invoicedAmount || +_case.fees || 0;
                    invDetails.cases[_cid].isCompleted = false;
                    invDetails.cases[_cid].receivedAmount = 0;
                    invDetails.cases[_cid].pendingAmount = invDetails.cases[_cid].invoicedAmount;
                }));

                invDetails.completedCaseIds = [];
                invDetails.payments = []; // Remove payments
                invDetails.accounting = []; // Remove accounting entries
                invDetails.receivedAmount = 0;
                invDetails.pendingAmount = +invDetails.amount;
                invDetails.paymentStatus = "PENDING";
                console.log("invoice updated: ", invDetails._id)
                return invDetails.save();
            }));

            // 3. update cases as paymentStatus pending
            await Promise.all(result.cases.map(async (_case, _cid) => {
                const caseID = _case.toString();

                // Update case data to reset payment status
                let newCaseData = { $set: { paymentIds: [] } };
                newCaseData = { ...newCaseData, paymentStatus: "PENDING" };
                console.log("case udpated: ", caseID)
                return await CaseServices.updateCase(caseID, newCaseData);
            }));

            return res.status(200).send({ status: true });

        }

        return res.status(200).send({ status: true, ...result })
    }
    catch (err) {
        console.log("=> error in resetcaseinvoicepayment: ", err);
        logger.error(`🚀 ~ file: Payment.controller.js:659 ~ module.exports.reversePayment= ~ err: ${err.toString()}`)
        return res.status(500).send({ status: false, error: err.toString() });
    }
};
