const moment = require('moment');
const { ObjectID } = require("mongodb");
const fs = require('fs');
const { logger } = require('../logger');
const Case = require('../Model/Case');
const Task = require('../Model/Task');
const Payments = require('../Model/Payment');
const Invoices = require('../Model/nInvoice');
const { workType } = require('../Types/Task');
const { FIELDS_ARRAY } = require('../Types/Report');

const CaseServices = require('../Services/CreateServices')('Case');
const { deepFlattenObject, generateExcel } = require('../Utils/Utility');
const SaveReportServices = require('../Services/CreateServices')('SavedReport');
const nInvoiceServices = require('../Services/CreateServices')('nInvoice');
const { dateRange } = require('../Utils/Utility');

const separateBy = (ary, separator = ",") => {
    return (ary && ary.split(separator)) && ary?.split(',').map((id, index) => id?.trim()).filter((id, index) => !!id) || []
}

module.exports.dashboardCounts = async (req, res) => {
    try {
        let count = {
            lead: 0,
            open: 0,
            inProcess: 0,
            waiting: 0,
            paymentPending: 0,
            completed: 0
        };
        let { start, end, organizationId } = req.query;
        let organizationIds = separateBy(organizationId);

        const leadCount = await CaseServices.count({
            ...(organizationIds.length > 0 && { 'organization._id': { $in: organizationIds } }),
            type: 'LEAD', caseStatus: { $ne: "CANCELLED" },
        });
        leadCount && (count.lead = leadCount.count || 0)

        const paymentPendingCount = await CaseServices.count({
            ...(organizationIds.length > 0 && { 'organization._id': { $in: organizationIds } }),
            caseStatus: { $ne: "CANCELLED" },
            paymentStatus: { $ne: "RECEIVED" }
        });
        paymentPendingCount && (count.paymentPending = paymentPendingCount.count || 0)

        const completedCaseCount = await CaseServices.count({
            ...(organizationIds.length > 0 && { 'organization._id': { $in: organizationIds } }),
            type: 'CASE',
            caseStatus: { $in: ["COMPLETED", "DELIVERED"] },
            'createdAt': dateRange(start, end)
        });
        completedCaseCount && (count.completed = completedCaseCount.count || 0)

        return res.status(200).send({ status: true, count })
    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:55 ~ module.exports.dashboardCounts= ~ e:, ${JSON.stringify(e)}`)
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.paymentDashboard = async (req, res) => {
    try {
        let counts = {
            invoicedAmount: 0,
            paymentPending: 0,
            paymentReceived: 0,
            tdsAmount: 0
        };

        let { start, end } = req.query;
        const taxInvoiceIdsRes = await Invoices.aggregate([
            { '$match': { invoiceStatus: 'COMPLETED', invoiceType: { $in: ["PROFORMA", "ADV_PROFORMA"] }, taxInvoiceId: { $exists: true }, paymentStatus: { $ne: "RECEIVED" } } },
            { $project: { taxInvoiceId: 1 } }
        ]);
        const taxInvoiceIds = taxInvoiceIdsRes.map(inv => inv.taxInvoiceId) || [];

        const invoicedAmount = await Invoices.aggregate([
            {
                '$match': {
                    invoiceStatus: 'COMPLETED',
                    'updatedAt': dateRange(start, end),
                    taxInvoiceId: { $exists: false }
                    // _id: { $nin: taxInvoiceIds }
                }
            },
            { '$group': { '_id': null, 'count': { $sum: '$amount' } } }
        ]);
        (invoicedAmount && invoicedAmount[0]) && (counts['invoicedAmount'] = invoicedAmount[0].count ? invoicedAmount[0].count : 0)

        const paymentPending = await Invoices.aggregate([
            {
                $match: {
                    invoiceStatus: "COMPLETED",
                    paymentStatus: { $ne: "RECEIVED" },
                    _id: { $nin: taxInvoiceIds }
                }
            },
            { $addFields: { pendingAmount: { $subtract: ["$amount", "$receivedAmount"] } } },
            { $group: { _id: null, count: { $sum: "$pendingAmount" } } }
        ]);
        (paymentPending && paymentPending[0]) && (counts['paymentPending'] = paymentPending[0].count ? paymentPending[0].count : 0)

        const paymentReceived = await Payments.aggregate([
            { '$match': { paymentStatus: 'COMPLETED' } },
            { '$group': { '_id': null, 'count': { $sum: '$amount' } } }
        ]);
        (paymentReceived && paymentReceived[0]) && (counts['paymentReceived'] = paymentReceived[0].count ? paymentReceived[0].count : 0)

        return res.status(200).send({ status: true, counts });
    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:55 ~ module.exports.dashboardCounts= ~ e:, ${JSON.stringify(e)}`)
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.usersMyTaskCount = async (req, res) => {

    const user = req.local.user;
    const role = user.role;
    const userId = req.query.userId || user._id.toString();
    let { start, end, organizationId } = req.query;

    let organizationIds = separateBy(organizationId).map((org) => ObjectID(org));

    let query = {
        status: "MY_TASK",
        ...((!!start && !!end) && { 'createdAt': dateRange(start, end) }),
        ...(organizationIds.length > 0 && { 'organizationId': { $in: organizationIds } }),
    };

    try {
        const statusWiseCaseCount = await Task.aggregate([
            { '$match': query },
            { $unwind: { path: "$users" } },
            {
                $group: {
                    _id: { userId: "$users.userId" },
                    user: { $first: "$users" },
                    count: { $sum: 1 },
                    tasks: {
                        $addToSet: {
                            id: "$_id",
                            workType: "$workType",
                            status: "$status",
                            duration: "$duration",
                        },
                    },
                },
            },
            {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [
                            {
                                'label': {
                                    $concat: [
                                        "$user.firstName",
                                        " ",
                                        "$user.lastName",
                                    ]
                                },
                            },
                            { userId: "$user.userId" },
                            { data: "$tasks" },
                            { 'count': '$count' }
                        ]
                    }
                }
            },
            { '$sort': { 'count': -1 } }
        ]);
        return res.status(200).send({ status: true, data: statusWiseCaseCount })
    }
    catch (e) {
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.workTypeOpenTaskCount = async (req, res) => {

    const user = req.local.user;
    const role = user.role;
    const userId = req.query.userId || user._id.toString();
    let { start, end, organizationId } = req.query;

    let organizationIds = separateBy(organizationId).map((org) => ObjectID(org));

    let query = {
        status: "OPEN",
        ...((!!start && !!end) && { 'createdAt': dateRange(start, end) }),
        ...(organizationIds.length > 0 && { 'organizationId': { $in: organizationIds } }),
    };

    try {
        const statusWiseCaseCount = await Task.aggregate([
            { '$match': query },
            {
                $group: {
                    _id: { workType: "$workType" },
                    count: { $sum: 1 },
                    tasks: {
                        $addToSet: {
                            id: "$_id",
                            workType: "$workType",
                            status: "$status",
                            duration: "$duration",
                        },
                    },
                },
            },
            {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [
                            {
                                'label': "$_id.workType",
                            },
                            { data: "$tasks" },
                            { 'count': '$count' }
                        ]
                    }
                }
            },
            { '$sort': { 'count': -1 } }
        ]);
        return res.status(200).send({ status: true, data: statusWiseCaseCount })
    }
    catch (e) {
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.statusWiseReportCount = async (req, res) => {

    const user = req.local.user;
    const role = user.role;
    const userId = req.query.userId || user._id.toString();
    let { start, end, organizationId } = req.query;

    let organizationIds = separateBy(organizationId);

    let query = {
        ...((!!start && !!end) && { 'createdAt': dateRange(start, end) }),
        ...(organizationIds.length > 0 && { 'organization._id': { $in: organizationIds } }),
    };

    if (req.query.userId || (!req.query.userId && !['VALUER', 'ADMIN', 'FRONTDESK', 'PRINTING', 'ACCOUNTANT'].includes(role))) {
        query['$or'] = [
            { "createdBy": userId },
            { "staff.pickupPerson._id": userId },
            { "staff.rateVerifier._id": userId },
            { "staff.siteEngineer._id": userId },
            { "staff.maker._id": userId },
            { "staff.checker._id": userId },
            { "staff.valuers._id": userId },
            { "staff.deliveryPerson._id": userId },
        ]
    }

    try {
        const statusWiseCaseCount = await Case.aggregate([
            { '$match': query },
            {
                '$group': {
                    '_id': { 'caseStatus': '$caseStatus' },
                    'caseCount': { '$sum': 1 },
                    'caseNumber': { '$addToSet': '$caseNumber' },
                }
            },
            {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [{ 'label': '$_id.caseStatus' }, { 'count': '$caseCount' }]
                    }
                }
            },
            { '$sort': { 'count': -1 } }
        ]);
        return res.status(200).send({ status: true, data: statusWiseCaseCount })
    }
    catch (e) {
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.clientWiseReportCount = async (req, res) => {
    try {

        let { start, end, organizationId } = req.query;
        let organizationIds = separateBy(organizationId);

        const clientWiseCaseCount = await Case.aggregate([
            {
                '$match': {
                    'createdAt': dateRange(start, end),
                    ...(organizationIds.length > 0 && { 'organization._id': { $in: organizationIds } }),
                }
            },
            {
                '$group': {
                    '_id': { 'client': '$client' },
                    'caseCount': { '$sum': 1 },
                    'caseNumber': {
                        '$addToSet': '$caseNumber'
                    },
                }
            }, {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [{ 'label': { $concat: ['$_id.client.displayName', '$_id.client.initial'] }, 'data': '$_id.client', 'caseNumber': '$caseNumber' }, { 'count': '$caseCount' }]
                    }
                }
            },
            { '$sort': { 'count': -1 } }
        ]);

        return res.status(200).send({ status: true, data: clientWiseCaseCount })
    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:49 ~ module.exports.clientWiseCaseCount= ~ e:, ${JSON.stringify(e)}`)
        return res.status(500).send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.getTaskCountByWorkType = async (req, res) => {
    try {
        let { start, end, organizationId } = req.query;
        let organizationIds = separateBy(organizationId).map((org) => ObjectID(org));

        const tasksByWorkTypeRes = await Task.aggregate([
            {
                '$match': {
                    'status': 'COMPLETED',
                    'completedAt': dateRange(start, end),
                    ...(organizationIds.length > 0 && { 'organizationId': { $in: organizationIds } }),
                },
            },
            {
                '$group': {
                    '_id': {
                        'workType': '$workType',
                    },
                    'count': { '$sum': 1 },
                },
            },
            {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [{ 'label': '$_id.workType', 'data': '$_id.client', 'caseNumber': '$caseNumber' }, { 'count': '$count' }]
                    }
                }
            },
            {
                '$sort': {
                    count: -1
                }
            }
        ]);

        return res.status(200).send({ status: true, data: tasksByWorkTypeRes || [] })
    }
    catch (e) {
        return res.send({ status: false, error: JSON.stringify(e) })
    }
};

module.exports.getTaskCountByUser = async (req, res) => {
    try {
        let { start, end } = req.query;

        const tasksByWorkTypeRes = await Task.aggregate([
            {
                '$match': {
                    'status': 'COMPLETED',
                    'completedAt': dateRange(start, end)
                },
            },
            {
                'lookup': {
                    'from': 'users',
                    'localField': 'users.userId',
                    'foreignField': '_id',
                    'as': 'user'
                }
            },
            { '$unwind': '$user' },
            {
                '$group': {
                    '_id': {
                        'user': '$user',
                        'userId': "$user._id"
                        // 'firstName': '$users.firstName',
                        // 'lastName': '$users.lastName'
                    },
                    'tasks': {
                        '$addToSet': {
                            'id': '$_id',
                            'workType': '$workType'
                        },
                    },
                    'count': { '$sum': 1 },
                }
            },
            {
                '$sort': {
                    count: 1
                }
            }
        ]);

        return res.status(200).send({ status: true, data: tasksByWorkTypeRes || [] })
    }
    catch (e) {
        return res.send({ status: false, error: JSON.stringify(e) })
    }
};

module.exports.staffWiseReportCount = async (req, res) => {
    try {

        let { start, end, organizationId } = req.query;
        let organizationIds = separateBy(organizationId);

        const clientWiseCaseCount = await Case.aggregate([
            {
                '$match': {
                    'createdAt': dateRange(start, end),
                    ...(organizationIds.length > 0 && { 'organization._id': { $in: organizationIds } }),
                }
            },
            {
                '$lookup': {
                    'from': 'users',
                    'localField': 'createdBy',
                    'foreignField': '_id',
                    'as': 'user'
                }
            },
            {
                '$unwind': '$user'
            },
            {
                '$group': {
                    '_id': { 'staff': '$user' },
                    'caseCount': { '$sum': 1 },
                    'caseNumber': {
                        '$addToSet': '$caseNumber'
                    },
                }
            }, {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [{ 'label': { $concat: ['$_id.staff.firstName', " ", '$_id.staff.lastName'] }, 'data': '$_id.staff', 'caseNumber': '$caseNumber' }, { 'count': '$caseCount' }]
                    }
                }
            },
            { '$sort': { 'count': -1 } }
        ]);

        return res.status(200).send({ status: true, data: clientWiseCaseCount })
    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:49 ~ module.exports.clientWiseCaseCount= ~ e:, ${JSON.stringify(e)}`)
        return res.status(500).send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.clientWiseInvoice = async (req, res) => {
    const { user } = req.local;
    const { clientId, branchId, organizationId, start, end } = req.query;
    let organizationIds = separateBy(organizationId).map((org) => ObjectID(org));

    let query = {
        invoiceStatus: "COMPLETED",
        taxInvoiceId: { $exists: false },
        ...(organizationIds.length > 0 && { 'organization._id': { $in: organizationIds } }),
        ...((!!start && !!end) && { 'updatedAt': dateRange(start, end) })
    }

    if (clientId) { query = { ...query, 'client._id': ObjectID(clientId) } }
    if (branchId) { query = { ...query, 'branch._id': ObjectID(branchId) } }

    try {
        const casesResult = await nInvoiceServices.aggregate([
            {
                $match: query
            },
            {
                '$group': {
                    '_id': { 'client': '$client' },
                    'amount': { '$sum': '$amount' },
                    'caseNumber': {
                        '$addToSet': '$caseNumber'
                    },
                }
            },
            {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [{ 'label': { $concat: ['$_id.client.displayName', "-", '$_id.client.initial'] }, 'data': '$_id.client', 'caseNumber': '$caseNumber' }, { 'count': '$amount' }]
                    }
                }
            },
            {
                $sort: {
                    count: -1
                }
            },
        ]);

        return res.send({ status: true, data: casesResult.data });
    } catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:392 ~ module.exports.clientWiseInvoice= ~ e, ${JSON.stringify(e)}`)
        return res.send({ status: false, error: JSON.stringify(e) });
    }
};

module.exports.staffWiseInvoice = async (req, res) => {
    const { user } = req.local;
    const { clientId, branchId, organizationId, start, end } = req.query;
    let organizationIds = separateBy(organizationId).map((org) => ObjectID(org));

    let query = {
        invoiceStatus: "COMPLETED",
        taxInvoiceId: { $exists: false },
        ...(organizationIds.length > 0 && { 'organization._id': { $in: organizationIds } }),
        ...((!!start && !!end) && { 'updatedAt': dateRange(start, end) })
    }

    if (clientId) { query = { ...query, 'client._id': ObjectID(clientId) } }
    if (branchId) { query = { ...query, 'branch._id': ObjectID(branchId) } }

    try {
        const casesResult = await nInvoiceServices.aggregate([
            {
                $match: {
                    ...query
                }
            },
            {
                $addFields: {
                    cases: { $first: "$cases" }
                }
            },
            {
                $lookup: {
                    from: "cases",
                    localField: "cases._id",
                    foreignField: "_id",
                    as: "cases",
                },
            },
            {
                $group: {
                    _id: {
                        staff: "$cases.createdBy",
                    },
                    amount: {
                        $sum: "$amount",
                    },
                    caseNumber: {
                        $addToSet: {
                            $first: "$cases.caseNumber",
                        },
                    },
                },
            },
            {
                $project: {
                    user: { $first: "$_id.staff" },
                    amount: 1,
                    caseNumber: 1,
                },
            },
            {
                $lookup: {
                    from: "users",
                    localField: "user",
                    foreignField: "_id",
                    as: "user",
                },
            },
            {
                $addFields: {
                    staff: {
                        $first: "$user",
                    },
                },
            },
            {
                $replaceRoot: {
                    newRoot: {
                        $mergeObjects: [
                            {
                                label: {
                                    $concat: [
                                        "$staff.firstName",
                                        " ",
                                        "$staff.lastName",
                                    ],
                                },
                                data: "$staff",
                                caseNumber: "$caseNumber",
                            },
                            {
                                count: "$amount",
                            },
                        ],
                    },
                },
            },
            {
                $sort: {
                    count: -1
                }
            },
        ]);

        return res.send({ status: true, data: casesResult.data });
    } catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:496 ~ module.exports.staffWiseInvoice= ~ e, ${JSON.stringify(e)}`)
        return res.send({ status: false, error: JSON.stringify(e) });
    }
};

module.exports.inactiveCaseList = async (req, res) => {
    try {

        const user = req.local.user;

        let query = [
            {
                $match: { "clientType.value": "BANK" },
            },
            {
                $sort: { createdAt: -1 }
            },
            {
                $group: {
                    _id: {
                        branchId: { $toObjectId: "$branch._id" },
                        clientId: { $toObjectId: "$client._id" }
                    },
                    case: { $first: "$$ROOT" },
                },
            },
            {
                $sort: { "case.createdAt": 1 },
            },
            {
                $lookup: { from: "clients", localField: "_id.clientId", foreignField: "_id", as: "client" },
            },
            {
                $lookup: { from: "branches", localField: "_id.branchId", foreignField: "_id", as: "branch" },
            },
            {
                $set: {
                    client: { $first: "$client" },
                    branch: { $first: "$branch" },
                },
            }
        ];

        const result = await Case.aggregate(query);
        return res.send({ status: false, clients: result });

    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:60 ~ module.exports.statusWiseLeadCount= ~ e, ${JSON.stringify(e)}`)
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.statusWiseCaseCount = async (req, res) => {

    const user = req.local.user;
    const role = user.role;
    const userId = req.query.userId || user._id.toString();
    let { start, end } = req.query;

    let query = {
        'type': "CASE",
        ...((!!start && !!end) && { 'createdAt': dateRange(start, end) })
    };

    if (req.query.userId || (!req.query.userId && !['VALUER', 'ADMIN', 'FRONTDESK', 'PRINTING', 'ACCOUNTANT'].includes(role))) {
        query['$or'] = [
            { "createdBy": userId },
            { "staff.pickupPerson._id": userId },
            { "staff.rateVerifier._id": userId },
            { "staff.siteEngineer._id": userId },
            { "staff.maker._id": userId },
            { "staff.checker._id": userId },
            { "staff.valuers._id": userId },
            { "staff.deliveryPerson._id": userId },
        ]
    }

    try {
        const statusWiseCaseCount = await Case.aggregate([
            { '$match': query },
            {
                '$group': {
                    '_id': { 'caseStatus': '$caseStatus' },
                    'caseCount': { '$sum': 1 },
                    'caseNumber': { '$addToSet': '$caseNumber' },
                }
            },
            {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [{ 'caseStatus': '$_id.caseStatus' }, { 'count': '$caseCount' }]
                    }
                }
            },
            { '$sort': { 'count': -1 } }
        ]);
        return res.status(200).send({ status: true, statusWiseCaseCount })
    }
    catch (e) {
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.userWiseTask = async (req, res) => {
    try {
        const { workType, searchText, start, end, orgBranchId, clientId, branchId, userId } = req.query;

        let query = {
            // ...((!!start && !!end) && { 'createdAt': dateRange(start, end) })
        };
        if (workType) { query = { workType }; }
        if (orgBranchId) { query = { organizationId: orgBranchId }; }

        if (!!searchText || searchText != "") {
            query = {
                ...query,
                '$or': [
                    { 'users.firstName': { $regex: searchText, $options: 'i' } },
                    { 'users.lastName': { $regex: searchText, $options: 'i' } }
                ]
            };
        }

        const totalCount = await Task.aggregate([
            { '$match': query },
            { '$sortByCount': "$status", }
        ])

        const userWiseTasksRes = await Task.aggregate([
            {
                '$match': {
                    'status': { '$ne': 'OPEN' },
                    'users.0': { $exists: true },
                    ...query
                }
            },
            {
                $unwind: { path: "$users" },
            },
            {
                $group: {
                    _id: { userId: "$users.userId" },
                    user: { $first: "$users" },
                    tasks: {
                        $addToSet: {
                            id: "$_id",
                            workType: "$workType",
                            status: "$status",
                            duration: "$duration",
                        },
                    },
                },
            },
            {
                $project: {
                    _id: false,
                    user: true,
                    totalTasks: { $size: "$tasks" },
                    COMPLETED: {
                        $size: {
                            $filter: {
                                input: "$tasks",
                                as: "task",
                                cond: {
                                    $and: [
                                        { $eq: ["$$task.status", "COMPLETED"] }
                                    ],
                                },
                            },
                        },
                    },
                    MY_TASK: {
                        $size: {
                            $filter: {
                                input: "$tasks",
                                as: "task",
                                cond: {
                                    $and: [
                                        { $eq: ["$$task.status", "MY_TASK"] }
                                    ],
                                },
                            },
                        },
                    },
                    REVIEW: {
                        $size: {
                            $filter: {
                                input: "$tasks",
                                as: "task",
                                cond: {
                                    $and: [
                                        { $eq: ["$$task.status", "REVIEW"] }
                                    ],
                                },
                            },
                        },
                    },
                    CANCELLED: {
                        $size: {
                            $filter: {
                                input: "$tasks",
                                as: "task",
                                cond: {
                                    $and: [
                                        { $eq: ["$$task.status", "CANCELLED"] }
                                    ],
                                },
                            },
                        },
                    },
                },
            },
            {
                '$sort': { 'user.firstName': 1 }
            }
        ]);
        return res.status(200).send({ status: true, userWiseTask: userWiseTasksRes || [], totalCount })
    }
    catch (e) {
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.userWiseTaskCount = async (req, res) => {
    try {
        let { start, end } = req.query;

        const userWiseTasksRes = await Task.aggregate([
            {
                '$match': {
                    'status': {
                        '$ne': 'OPEN'
                    },
                    'workType': { '$ne': 'CLIENT_VISIT' },
                    'users.0': { $exists: true },
                    'createdAt': dateRange(start, end)
                }
            }, {
                '$lookup': {
                    'from': 'users',
                    'localField': 'users.userId',
                    'foreignField': '_id',
                    'as': 'user'
                }
            },
            {
                '$group': {
                    '_id': {
                        'user': '$user',
                        'userId': "$user._id"
                        // 'firstName': '$users.firstName',
                        // 'lastName': '$users.lastName'
                    },
                    'tasks': {
                        '$addToSet': {
                            'id': '$_id',
                            'workType': '$workType',
                            'status': '$status',
                            'duration': '$duration'
                        }
                    }
                }
            },
            { '$unwind': '$_id.user' },
            { '$sort': { '_id.user.firstName': 1 } }
        ]);
        return res.status(200).send({ status: true, userWiseTaskCount: userWiseTasksRes || [] })
    }
    catch (e) {
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

const getTaskCountsByWorkType = async (workType, query = {}) => {
    return await Task.aggregate([
        {
            '$match': {
                'status': { '$ne': 'OPEN' }, 'workType': workType,
                ...query
            },
        },
        {
            '$lookup': { 'from': 'users', 'localField': 'users.userId', 'foreignField': '_id', 'as': 'user' },
        },
        {
            '$unwind': '$user',
        },
        {
            '$group': {
                '_id': {
                    'firstName': '$user.firstName',
                    'lastName': '$user.lastName',
                },
                'count': { '$sum': 1 },
            },
        },
        { '$sort': { 'count': -1 } }
    ]);
};

module.exports.topUsers = async (req, res) => {
    let { start, end, organizationId } = req.query;

    let organizationIds = separateBy(organizationId).map((org) => ObjectID(org));

    let query = {
        'completedAt': dateRange(start, end),
        ...(organizationIds.length > 0 && { 'organizationId': { $in: organizationIds } }),
    }

    try {
        const maker = await getTaskCountsByWorkType(workType.MAKER, query);
        const checker = await getTaskCountsByWorkType(workType.CHECKER, query);
        const pickup = await getTaskCountsByWorkType(workType.LEAD, query);
        const siteEngineer = await getTaskCountsByWorkType(workType.SITE_INSPECTION, query);
        const rateVerifier = await getTaskCountsByWorkType(workType.RATE_VERIFY, query);

        return res.send({ status: true, maker, checker, pickup, siteEngineer, rateVerifier });
    } catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:583 ~ module.exports.topUsers= ~ e, ${JSON.stringify(e)}`)
        return res.send({ status: false });
    }
}

module.exports.clientWiseCaseCount = async (req, res) => {
    try {

        let { start, end } = req.query;

        const clientWiseCaseCount = await Case.aggregate([
            {
                '$match': {
                    'type': "CASE",
                    // 'clientType.label': { $nin: ["Private Client", 'Private Clients'] },
                    'createdAt': dateRange(start, end)
                }
            },
            {
                '$group': {
                    '_id': { 'client': '$client' },
                    'caseCount': { '$sum': 1 },
                    'caseNumber': {
                        '$addToSet': '$caseNumber'
                    },
                }
            }, {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [{ 'client': '$_id.client' }, { 'count': '$caseCount', caseNumber: '$caseNumber' }]
                    }
                }
            },
            { '$sort': { 'count': -1 } }
        ]);

        return res.status(200).send({ status: true, clientWiseCaseCount })
    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:49 ~ module.exports.clientWiseCaseCount= ~ e:, ${JSON.stringify(e)}`)
        return res.status(500).send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.clientWiseTaskCount = async (req, res) => {
    try {
        let { start, end } = req.query;

        const clientWiseTaskCount = await Case.aggregate([
            {
                '$match': {
                    'createdAt': dateRange(start, end)
                }
            },
            {
                $lookup: {
                    from: "cases",
                    localField: "caseId",
                    foreignField: "_id",
                    as: "case",
                }
            },
            {
                $unwind: { path: "$case" },
            },
            {
                $group: {
                    _id: { client: "$case.client", workType: "$workType" },
                    taskCount: { $sum: 1 },
                    caseNumber: { $addToSet: "$case.caseNumber" }
                }
            },
            {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [{ 'client': '$_id.client' }, { 'count': '$caseCount', caseNumber: '$caseNumber' }]
                    }
                }
            },
            { '$sort': { 'count': -1 } }
        ]);

        return res.status(200).send({ status: true, clientWiseTaskCount })
    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:104 ~ module.exports.clientWiseTaskCount= ~ e:, ${JSON.stringify(e)}`)
        return res.status(500).send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.statusWiseLeadCount = async (req, res) => {
    try {

        const user = req.local.user;
        const role = user.role;
        const userId = "" + (req.query.userId || user._id);
        let query = {};

        if (req.query.userId || (!req.query.userId && !['VALUER', 'ADMIN', 'FRONTDESK', 'PRINTING', 'ACCOUNTANT'].includes(role))) {
            query['$or'] = [
                { "createdBy": userId },
                { "staff.pickupPerson._id": userId }
            ]
        }

        const leadCount = await CaseServices.count({ caseStatus: "LEAD", type: "LEAD", ...query });
        if (!leadCount.status || leadCount.error) return res.status(500).send(leadCount);

        const approximateLeadCount = await CaseServices.count({ type: "APPROXIMATE_LEAD", caseStatus: { $ne: "CANCELLED" }, ...query });
        if (!approximateLeadCount.status || approximateLeadCount.error) return res.status(500).send(approximateLeadCount);

        const cancelledLeadCount = await CaseServices.count({ caseStatus: "CANCELLED", type: { $in: ["LEAD", "APPROXIMATE_LEAD"] }, ...query });
        if (!cancelledLeadCount.status || cancelledLeadCount.error) return res.status(500).send(cancelledLeadCount);

        const inquiryCount = await CaseServices.count({ type: "INQUIRY_LEAD", caseStatus: { $ne: "CANCELLED" }, ...query });
        if (!inquiryCount.status || inquiryCount.error) return res.status(500).send(inquiryCount);

        const quotationCount = await CaseServices.count({ type: "QUOTATION_LEAD", caseStatus: { $ne: "CANCELLED" }, ...query });
        if (!quotationCount.status || quotationCount.error) return res.status(500).send(quotationCount);

        const statusWiseLeadCount = [
            { caseStatus: "LEAD", count: leadCount.count },
            { caseStatus: "APPROXIMATE_LEAD", count: approximateLeadCount.count },
            { caseStatus: "INQUIRY_LEAD", count: inquiryCount.count },
            { caseStatus: "QUOTATION_LEAD", count: quotationCount.count },
            { caseStatus: "CANCELLED", count: cancelledLeadCount.count }
        ]
        return res.status(200).send({ status: true, statusWiseLeadCount })

    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:60 ~ module.exports.statusWiseLeadCount= ~ e:, ${JSON.stringify(e)}`)
        return res.send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.compareWithTask = async (req, res) => {
    try {
        const user = req.local.user;
        let { start, end } = req.query;

        const data = await Task.aggregate([
            {
                '$match': {
                    'status': {
                        '$ne': 'OPEN'
                    },
                    'users.0': { $exists: true },
                    'users.userId': ObjectID(user._id.toString()),
                    'createdAt': dateRange(start, end)
                }
            },
            {
                '$group': {
                    '_id': {
                        'workType': '$workType',
                    },
                    'tasks': {
                        '$addToSet': {
                            'id': '$_id',
                            'status': '$status',
                            'duration': '$duration'
                        }
                    }
                }
            },
            { '$unwind': '$_id.user' },
            { '$sort': { '_id.user.firstName': 1 } }
        ]);

        console.log(`🚀 ~ module.exports.compareWithTask= ~ userWiseTasksRes:`, data);
        return res.status(200).send({ status: true, data: data || [] })
    }
    catch (e) {
        return res.status(500).send({ status: false, error: JSON.stringify(e) })
    }
}

module.exports.clientVisitCounts = async (req, res) => {
    try {
        const user = req.local.user;
        let { start, end } = req.query;
        let query = {
            'status': 'COMPLETED',
            'workType': "CLIENT_VISIT",
            'users.0': { $exists: true },
            // 'users.userId': ObjectID(user._id.toString()),
            'createdAt': dateRange(start, end)
        }

        const totalCount = await Task.find(query).count();

        const data = await Task.aggregate([
            {
                '$match': query
            },
            {
                '$lookup': { 'from': 'users', 'localField': 'users.userId', 'foreignField': '_id', 'as': 'user' },
            },
            {
                '$unwind': '$user',
            },
            {
                '$group': {
                    '_id': {
                        '_id': '$user._id',
                        'firstName': '$user.firstName',
                        'lastName': '$user.lastName',
                    },
                    'count': { '$sum': 1 },
                }
            },
            {
                '$replaceRoot': {
                    'newRoot': {
                        '$mergeObjects': [
                            {
                                'label': {
                                    $concat: [
                                        "$_id.firstName",
                                        " ",
                                        "$_id.lastName",
                                    ]
                                }
                            },
                            { 'userId': "$_id._id" },
                            { 'count': '$count' }
                        ]
                    }
                }
            },
            { '$sort': { 'count': -1 } }
        ]);

        return res.status(200).send({ status: true, data: data || [], totalCount: totalCount || 0 })
    }
    catch (e) {
        return res.status(500).send({ status: false, error: e.toString() })
    }
}

module.exports.customReport = async (req, res) => {
    try {

        let { download = "false", reportOf = "Case", dateRangeOn = "createdAt", start, end, organizationId, clientType, client, branch, caseStatus, workType, taskStatus, userId, display = null } = req.query;
        const modelName = reportOf?.toLowerCase() || "";
        const Models = { task: Task, case: Case };
        if (!Models[modelName]) return res.status(500).send({ status: false, error: "Invalid request!" });
        const headers = FIELDS_ARRAY[reportOf.toUpperCase()] || [];

        let organizationIds = separateBy(organizationId);
        let clientIds = separateBy(client);
        let branchIds = separateBy(branch);
        let workTypes = separateBy(workType);
        let caseStatuses = separateBy(caseStatus).map((cs, id) => cs.toUpperCase());
        let workStatus = separateBy(taskStatus).map((cs, id) => cs.toUpperCase());

        let query = [];
        let project = {};
        display = display || headers
        display.map((f, id) => { project[f] = true });

        switch (reportOf.toLowerCase()) {
            case "task":
                query = [
                    {
                        '$match': {
                            ...((start && end) && { [dateRangeOn]: dateRange(start, end) }),
                            ...(organizationIds.length > 0 && { 'organizationId': { $in: organizationIds } }),
                            ...(workTypes.length > 0 && { workType: { $in: workTypes } }),
                            ...(userId && { 'users.userId': ObjectID(userId) }),
                            ...(workStatus.length > 0 && { status: { $in: workStatus } })
                        }
                    },
                    {
                        '$lookup': {
                            from: "cases",
                            localField: 'caseId',
                            foreignField: '_id',
                            as: "cases"
                        }
                    },
                    { '$unwind': { path: '$cases' } },
                    {
                        '$match': {
                            ...(clientIds.length > 0 && { 'cases.client._id': { $in: clientIds } }),
                            ...(branchIds.length > 0 && { 'cases.branch._id': { $in: branchIds } }),
                            ...(clientType && { 'cases.clientType.value': clientType })
                        }
                    },
                    {
                        '$addFields': {
                            'pickup_person': {
                                '$concat': [
                                    "$case.staff.pickupPerson.firstName",
                                    " ",
                                    "$case.staff.pickupPerson.lastName",
                                ],
                            },
                        }
                    },
                    { '$project': { 'case': false } },
                    { '$sort': { [dateRangeOn]: -1 } }
                ];
                break;
            case 'case':
                query = [
                    {
                        '$match': {
                            ...((start && end) && { [dateRangeOn]: dateRange(start, end) }),
                            ...(organizationIds.length > 0 && { 'organization._id': { $in: organizationIds } }),
                            ...(caseStatus.length > 0 && { caseStatus: { $in: caseStatuses } }),
                            ...(clientType && { 'clientType.value': clientType }),
                            ...(clientIds.length > 0 && { 'client._id': { $in: clientIds } }),
                            ...(branchIds.length > 0 && { 'branch._id': { $in: branchIds } }),
                            ...(userId && {
                                '$or': [
                                    { "createdBy": userId },
                                    { "staff.pickupPerson._id": userId },
                                    { "staff.maker._id": userId },
                                    { "staff.checker._id": userId },
                                    { "staff.valuers._id": userId },
                                    { "staff.rateVerifier._id": userId },
                                    { "staff.siteEngineer._id": userId }
                                ]
                            })
                        }
                    },
                    { '$sort': { [dateRangeOn]: -1 } }
                ]
                break;
            default:
                break;
        }

        const reportCountRes = await Models[modelName].aggregate([...query, { '$count': 'count' }]).allowDiskUse(true);
        const totalCount = (reportCountRes && reportCountRes[0] && reportCountRes[0].count) && reportCountRes[0].count || 0
        query = [...query, { '$project': project }];

        if (!!download && download == "true") {
            const reportRes = await Models[modelName].aggregate(query).allowDiskUse(true);
            const data = await JSON.parse(JSON.stringify(reportRes));
            const flattedData = await deepFlattenObject(data);
            const filePath = await generateExcel(flattedData, `${modelName}-report.xlsx`);

            return res.sendFile(filePath, {}, (err) => {
                if (err) {
                    logger.error('Error occurred while sending file:' + err.toString());
                    return res.status(err.status).end();
                } else {
                    fs.unlinkSync(filePath); // Remove the temporary file after sending
                }
            });
        }
        else {
            query = [
                ...query,
                { "$skip": req.local.startIndex },
                { "$limit": req.local.limit }
            ];
            const reportRes = await Models[modelName].aggregate(query).allowDiskUse(true);

            return res.status(200).send({ status: true, data: reportRes, headers: Object.keys(reportRes[0]), count: totalCount });
        }
    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:1021 ~ module.exports.customReport= ~ e, ${e.toString()}`)
        return res.status(500).send({ status: false, error: e.toString() })
    }
}

module.exports.getAllSavedReports = async (req, res) => {
    try {
        const SavedReportResult = await SaveReportServices.find({});

        if (!SavedReportResult.status || SavedReportResult.error) return res.status(500).send(SavedReportResult);

        return res.status(200).send(SavedReportResult);
    }
    catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:1080 ~ module.exports.getAllSavedReports= ~ e, ${JSON.stringify(e)}`)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
}

module.exports.getSavedReport = async (req, res) => {
    try {
        const id = req.params.id;

        const savedReportResult = await SaveReportServices.findById(id);
        if (!savedReportResult.status || savedReportResult.error) return res.status(500).send(savedReportResult);

        return res.status(200).send(savedReportResult);
    } catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:1094 ~ module.exports.getSavedReport= ~ e, ${JSON.stringify(e)}`)
        return res.status(500).send({ status: false, error: e });
    }
}

module.exports.createSaveReport = async (req, res) => {
    try {
        const user = req.local.user;
        let filters = req.local.savedReport || {};
        let data = { filters, title: filters.title }
        data.createdBy = user._id;
        // data.organizationId = user.organization._id;

        const saveReportResult = await SaveReportServices.create(data);
        if (!saveReportResult.status || saveReportResult.error) return res.status(500).send(saveReportResult);

        return res.status(200).send({ status: true, data: saveReportResult.data });

    } catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:1113 ~ module.exports.createSaveReport= ~ e, ${JSON.stringify(e)}`)
        return res.status(500).send({ status: false, error: e });
    }
}

module.exports.updateSaveReport = async (req, res) => {
    try {
        const user = req.local.user;

        const id = req.params.id;
        const filters = req.local.savedReport;
        let data = { filters, ...(filters.title && { title: filters.title }) };
        data.updatedBy = user._id;

        const updateSaveReportRes = await SaveReportServices.updateById(id, data);
        if (!updateSaveReportRes.status || updateSaveReportRes.error) return res.status(500).send(updateSaveReportRes);

        return res.status(200).send(updateSaveReportRes);
    } catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:1132 ~ module.exports.updateSaveReport= ~ e, ${JSON.stringify(e)}`)
        return res.status(500).send({ status: false, error: e });
    }
}

module.exports.deleteSavedReport = async (req, res) => {
    try {
        const id = req.params.id;

        const deleteSaveReportRes = await SaveReportServices.deleteById(id);
        if (!deleteSaveReportRes.status || deleteSaveReportRes.error) return res.status(500).send(deleteSaveReportRes);

        return res.status(200).send(deleteSaveReportRes);
    } catch (e) {
        logger.error(`🚀 ~ file: Report.controller.js:1146 ~ module.exports.deleteSaveReport= ~ e:, ${e.toString()}`)
        return res.status(500).send({ status: false, error: e });
    }
}