const RoleServices = require('../Services/Role.services');
const UserServices = require('../Services/User.services');
const { PermissionToBinary } = require('../Services/PermissionToBinaryServices');
/**
 * @description This is used for get list of all roles 
 * @param {*} req 
 * @param {*} res 
 * @returns Object RoleResult { status: true, roles: Array }
 * -  {status : false , error : string }
 */
module.exports.getAllRoles = async (req, res) => {
    try {
        const role = req.local.user.role;
        if (role == "SUPERADMIN") {
            const RoleResult = await RoleServices.getAllRoles({ status: true });
            if (!RoleResult.status || RoleResult.error) return res.status(500).send(RoleResult);

            return res.status(200).send(RoleResult);
        }
        const RoleResult = await RoleServices.getAllRoles({ status: true, name: { $nin: ["SUPERADMIN"] } });
        if (!RoleResult.status || RoleResult.error) return res.status(500).send(RoleResult);

        return res.status(200).send(RoleResult);
    }
    catch (e) {
        console.log("Error: ", e);
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};
/**
 * @description This is used for get one role
 * @param {*} req.params.id from that get RoleId
 * @param {*} res 
 * @returns Object RoleResult { status: true, roles: Object }
 * -  {status : false , error : string }
 */
module.exports.getRole = async (req, res) => {
    try {

        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const RoleResult = await RoleServices.getRole(id);
        if (!RoleResult.status || RoleResult.error) return res.status(500).send(RoleResult);

        return res.status(200).send(RoleResult);
    }
    catch (e) {
        console.log("Error: ", e);
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};
/**
 * @description This is used for create new role
 * @param {*} req.body get Object
 * @param {*} res 
 * @returns Object RoleResult { status: true, roles: Object }
 * -  {status : false , error : string }
 * - { status: false, error: "Already Created!" }
 */
module.exports.createRole = async (req, res) => {
    try {
        const data = req.body;

        data.organization = req.local.user.organization._id;

        const alredyCreated = await RoleServices.getAllRoles({ name: data.name })
        if (!alredyCreated.status || alredyCreated.error) return res.status(500).send(alredyCreated);

        if (alredyCreated.roles.length == 0) {

            if (data.permissions) {
                // ===== check for permissions object in request and convert to binary ie.{case:0101, roles: 0110} =====
                var res_result = await PermissionToBinary(data.permissions);
                data.permission = res_result.result;
            }

            const RoleResult = await RoleServices.createRole(data);
            if (!RoleResult.status || RoleResult.error) return res.status(500).send(RoleResult);

            return res.status(200).send(RoleResult);
        }
        else {
            return res.status(500).send({ status: false, error: "Already Created!" });
        }
    }
    catch (e) {
        console.log("Error: ", e);
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }

};
/**
 * @description This is used for update role
 * @param {*} req.params.id from that get RoleId and from req.body get Object
 * @param {*} res 
 * @returns Object RoleResult { status: true, roles: Object }
 * -  {status : false , error : string }
 */
module.exports.updateRole = async (req, res) => {

    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const data = req.body;
        const userUpdate = req.query.userUpdate;

        if (data.permissions) {
            // ===== check for permissions object in request and convert to binary ie.{case:0101, roles: 0110} =====
            var res_result = await PermissionToBinary(data.permissions);
            // console.log(res_result);
            data.permission = res_result.result;
        }

        const RoleResult = await RoleServices.editRole(id, data);
        if (!RoleResult.status || RoleResult.error) return res.status(500).send(RoleResult);

        if (userUpdate == 'true') {

            const UserResult = await UserServices.editMultipleUser({ role: RoleResult.roles.name }, { permission: RoleResult.roles.permission })
        }

        return res.status(200).send(RoleResult);

    }
    catch (e) {
        console.log("Error: ", e);
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};
/**
 * @description This is used for edit permissions of roles 
 * @param {*} req.params.id from that get RoleId and from req.body get Object
 * @param {*} res 
 * @returns Object RoleResult { status: true, roles: roles }
 * -  {status : false , error : string }
 */
module.exports.editRole = async (req, res) => {

    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const data = req.body;

        // console.log(data.branch.access.value)

        // data.organization = req.local.user.organization._id;
        // console.log(data.name);

        const role = await RoleServices.getRole({ _id: id });
        if (!role.status || role.error) return res.status(500).send(role);

        var rolePermission = role.roles.permissions;

        // let value = role.roles.permissions[data.name][data.permission].value
        // console.log(value);
        var name = rolePermission.map((value) => {
            // console.log(value.name)
            if (value.name == data.name) {
                value.value = data.value
            }
            return value.name;
        });
        // console.log(name)
        if (!name.includes(data.name)) {
            rolePermission.push(data);
        }
        const RoleResult = await RoleServices.editRole(id, { permissions: rolePermission });
        if (!RoleResult.status || RoleResult.error) return res.status(500).send(RoleResult);

        return res.status(200).send(RoleResult);

    }
    catch (e) {
        console.log("Error: ", e);
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

module.exports.deleteRole = async (req, res) => {
    // return res.status(200).send({ message: "Hi There Its  Admin" });
};


