const fs = require('fs');
const path = require('path');
const moment = require('moment');
const { ObjectID } = require('mongodb');
const _ = require('loadsh');
const TaskServices = require('../Services/Task.services');
const UserServices = require('../Services/User.services');
const RoleServices = require('../Services/Role.services');
const CaseServices = require('../Services/Case.services');
const ReportTemplateServices = require('../Services/CreateServices')('ReportTemplate');
const nInvoiceController = require('./nInvoice.controller');
const nInvoiceServices = require('../Services/CreateServices')('nInvoice');
const utilitiesController = require('../Controllers/Utilities.controller');
const creditOfficerServices = require('../Services/CreateServices')('CreditOfficer');

const { logger } = require('../logger');
const CONFIG = require('../Constants/Config');
const { workType } = require('../Types/Task');
const { sendExpoNotification } = require('../Services/ExpoNotification.services');
const base64toFile = require('../Utils/Base64ToFile');
const { dateRange } = require('../Utils/Utility');

/**
 * @description This is used for get list of all tasks
 * @param {*} req 
 * @param {*} res 
 * @returns Object TaskResult { status: true, tasks: Array }
 * - { status : false, error : string }
 */
module.exports.getDashboardTaskCount = async (req, res) => {
    try {
        const user = req.local.user;
        const orgBranchPermission = user.orgBranches || [];
        const userId = req.query.userId || null;
        let responseCounts = {
            OPEN: { count: 0, workTypes: [] },
            MY_TASK: { count: 0, workTypes: [] },
            REVIEW: { count: 0, workTypes: [] },
            COMPLETED: { count: 0, workTypes: [] }
        }
        let query = { 'users.userId': ObjectID(userId ? userId : user._id) };

        const RoleRes = await RoleServices.findByRoleName({ name: user.role });
        if (!RoleRes.status) return res.status(500).send(RoleRes)

        let openQuery = {};
        let obWorkTypes = [];
        let obQuery = orgBranchPermission.map((ob, id) => {
            obWorkTypes = _.union(obWorkTypes, ob.workTypes);
            return { $and: [{ organizationId: ob._id }, { workType: { $in: ob.workTypes } }] };
        })
        openQuery = { status: "OPEN", ...(obQuery.length > 0 ? { $or: obQuery } : {}) }

        const OPEN = await TaskServices.count(openQuery);
        const MY_TASK = await TaskServices.count({ status: "MY_TASK", workType: { $ne: "CLIENT_VISIT" }, ...query });
        const REVIEW = await TaskServices.count({ status: "REVIEW", workType: { $ne: "CLIENT_VISIT" }, ...query });
        const COMPLETED = await TaskServices.count({ status: "COMPLETED", workType: { $ne: "CLIENT_VISIT" }, ...query });

        const counts = await TaskServices.aggregate([
            {
                $match: {
                    $or: [
                        openQuery,
                        { status: { $ne: "OPEN" }, ...query }
                    ],
                }
            },
            {
                $group: {
                    _id: { workType: "$workType", status: "$status" },
                    count: { $sum: 1 },
                }
            },
            {
                $group: {
                    _id: { status: "$_id.status" },
                    count: { $sum: "$$ROOT.count" },
                    workTypes: { $push: { workType: "$_id.workType", count: "$count" } },
                }
            },
            {
                $project: {
                    _id: false,
                    status: "$_id.status",
                    workTypes: 1,
                    count: 1,
                }
            }
        ]);

        const taskWorkTypeCounts = counts.tasks.map((st, id) => {
            if (!!responseCounts[st.status]) { responseCounts[st.status] = st; }
        });

        console.log("🚀 ~ file: Task.controller.js:96 ~ module.exports.getDashboardTaskCount= ~ responseCounts:", responseCounts)
        return res.status(200).send({ status: true, counts: responseCounts, count: { OPEN: OPEN.count || 0, MY_TASK: MY_TASK.count || 0, REVIEW: REVIEW.count || 0, COMPLETED: COMPLETED.count || 0 }, orgBranchWorkTypes: orgBranchPermission, workTypes: (obWorkTypes.length > 0 ? obWorkTypes : RoleRes.roles.taskWorkTypes) || [] });
    }
    catch (e) {
        logger.error("🚀 ~ file: Task.controller.js:48 ~ module.exports.getAllTasks= ~ e:", e.toString());
        return res.status(500).send({ status: false, error: e.toString() });
    }
};

/**
 * @description This is used for get list of all tasks
 * @param {*} req 
 * @param {*} res 
 * @returns Object TaskResult { status: true, tasks: Array }
 * - { status : false, error : string }
 */
module.exports.getAllTasks = async (req, res) => {
    try {
        let query = {};
        const searchText = req.query.searchText || "";
        let searchRegex = {};
        const user = req.local.user;
        const orgBranchPermission = user.orgBranches || [];
        const userId = req.query.userId || user._id;

        let obWorkTypes = [];
        orgBranchPermission.map((ob, id) => { obWorkTypes = _.union(obWorkTypes, ob.workTypes) })

        const status = req.query.status || 'OPEN';
        const workType = req.query.workType || null;
        let { latitude, longitude, startDate, endDate } = req.query;

        if (!!searchText) {
            searchRegex = {
                '$or': [
                    { 'caseId.caseNumber': { $regex: searchText, $options: 'i' } },
                    { 'caseId.ownerName': { $regex: searchText, $options: 'i' } },
                    { "caseId.typeOfReport.label": { $regex: searchText, $options: 'i' } },
                ]
            }
        }

        if ((!startDate || !endDate || !moment(startDate, "DD-MM-YYYY").isValid() || !moment(endDate, "DD-MM-YYYY").isValid())) {
            let date = new Date();
            endDate = new Date(date.getFullYear(), date.getMonth() + 1, 0);
            startDate = new Date(date.getFullYear(), date.getMonth(), 1);
        }
        else {
            startDate = new Date(moment(startDate, "DD-MM-YYYY").startOf('day'));
            endDate = new Date(moment(endDate, "DD-MM-YYYY").endOf('day'));
        }

        // const RoleRes = await RoleServices.findByRoleName({ name: user.role });
        // if (!RoleRes.status) return res.status(500).send(RoleRes)

        !!workType && (query.workType = { $in: [workType] });
        query.status = status;

        if (status == "OPEN") {
            let obQuery = orgBranchPermission.map((ob, id) => {
                return {
                    $and: [
                        { organizationId: ObjectID(ob._id) },
                        { workType: { $in: ob.workTypes } }
                    ]
                };
            })
            query = { ...query, ...(obQuery.length > 0 ? { $or: obQuery } : {}) }
        } else {
            query = { ...query, 'users.userId': ObjectID(userId) }
        }

        if (status == "COMPLETED") { query = { ...query, completedAt: { $gte: startDate, $lte: endDate } } }

        const TaskCountResult = await TaskServices.aggregate([
            { '$match': query },
            { '$lookup': { 'from': 'cases', 'localField': 'caseId', 'foreignField': '_id', 'as': 'caseId' } },
            { '$unwind': { 'path': '$caseId' } },
            { '$match': searchRegex },
            { '$count': "count" }
        ]);
        if (!TaskCountResult.status || TaskCountResult.error) return res.status(500).send(TaskCountResult);
        const taskCount = (TaskCountResult.tasks[0] && TaskCountResult.tasks[0].count) ? TaskCountResult.tasks[0].count : 0

        const TaskResult = await TaskServices.aggregate(
            [
                (
                    (latitude && longitude) ?
                        {
                            '$geoNear': {
                                near: {
                                    type: 'Point',
                                    coordinates: [+longitude, +latitude]
                                },
                                distanceField: "distance",
                                // maxDistance: 10000,
                                spherical: true, query: query,
                            }
                        }
                        : { '$match': query }
                ),
                { '$lookup': { 'from': 'cases', 'localField': 'caseId', 'foreignField': '_id', 'as': 'caseId' } },
                { '$unwind': { 'path': '$caseId' } },
                { '$lookup': { 'from': 'organizations', 'localField': 'organizationId', 'foreignField': '_id', 'as': 'organizationId' } },
                { '$unwind': { 'path': '$organizationId' } },
                { '$match': searchRegex },
                { '$sort': { "caseId.caseNumber": -1 }, },
                { "$skip": req.local.startIndex },
                { "$limit": req.local.limit },
            ]
        );

        if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);
        return res.status(200).send({ status: true, tasks: TaskResult.tasks, count: taskCount });
    }
    catch (e) {
        console.log("🚀 ~ file: Task.controller.js:48 ~ module.exports.getAllTasks= ~ e:", e)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used for get one task
 * @param {*} req.params.id from that get taskId  
 * @param {*} res 
 * @returns Object TaskResult { status: true, TaskResult, Empanelments: Object.values(arr) }
 * - { status : false, error : string }
 */
module.exports.getTask = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const TaskResult = await TaskServices.findById(id, { populate: ['organizationId', 'caseId'] });
        if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);

        const caseId = TaskResult.tasks.caseId;

        const caseTasksResult = await TaskServices.findAll({ caseId: caseId }, { populate: ['organizationId', 'caseId'] });
        if (!caseTasksResult.status || caseTasksResult.error) return res.status(500).send(TaskResult);

        const ReportResult = await ReportTemplateServices.find({ query: { "client._id": caseId.client._id, 'propertyType': caseId.propertyType.label, 'typeOfReport': caseId.typeOfReport.label, status: true } });
        if (!ReportResult.status || ReportResult.error) return res.status(500).send(ReportResult);

        const hasInvoiceRes = await nInvoiceServices.find({ query: { 'cases._id': TaskResult.tasks.caseId._id }, sort: { 'createdAt': 1 } });
        if (!hasInvoiceRes.status || hasInvoiceRes.error) return res.status(500).send(hasInvoiceRes);

        let task = { ...TaskResult.tasks.toJSON(), hasExternalReport: true, externalReportReason: "" };
        if (caseId.clientType && caseId.clientType.value == "PRIVATE_CLIENT") {
            task.hasExternalReport = false;
            task.externalReportReason = "PRIVATE CIENT";
        }
        else if (!!ReportResult.data && ReportResult.data.length > 0) {
            task.hasExternalReport = false;
            task.externalReportReason = "Template Not available";
        } 

        if (caseId.isExternalReport) {
            task.hasExternalReport = true;
            task.externalReportReason = "FOrce External Report";
        }
        task.invoice = hasInvoiceRes.data;

        // if (hasInvoiceRes.invoices.length == 0) {
        //     const newInvoiceRes = await InvoiceServices.createInvoice({ organization: TaskResult.tasks.caseId.organization, case: TaskResult.tasks.caseId, invoices: [{ invoiceType: "PROFORMA" }] })
        // }

        // const invoiceRes = await InvoiceServices.getAllInvoices({ 'case._id': TaskResult.tasks.caseId._id });
        // task.invoice = invoiceRes.invoices[0];

        return res.status(200).send({ status: true, tasks: task, caseTasks: caseTasksResult.tasks });
    }
    catch (e) {
        console.log("Error: ", e);
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used for create new task
 * @param {*} req.local.task from that get task Object
 * @param {*} res 
 * @returns Object TaskResult { status: true, tasks: object }
 * - { status : false, error : string }
 */
module.exports.createTaskRequest = async (req, res) => {
    try {
        const user = req.local.user;
        const data = req.local.task;

        data.organizationId = user.organization._id;

        const checkTaskResult = await TaskServices.findAll(data);
        if (!checkTaskResult.status || checkTaskResult.error) return res.status(500).send(checkTaskResult);

        const TaskResult = await TaskServices.create(data);
        if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);

        return res.status(200).send(TaskResult);
    }
    catch (err) {
        console.log(err)
        return res.status(500).send({ status: false, error: err });
    }
};

module.exports.createTask = async (data) => {
    try {
        const checkTaskResult = await TaskServices.findAll({ caseId: data.caseId, workType: data.workType }, {});
        if (!checkTaskResult.status || checkTaskResult.error) return checkTaskResult;
        if (checkTaskResult.count > 0) return { ...checkTaskResult, status: false };

        const TaskResult = await TaskServices.create(data);
        if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);

        return TaskResult
    }
    catch (err) {
        return { status: false, error: err };
    }
};

module.exports.ClientVisitTask = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const TaskResult = await TaskServices.findById(id, { populate: ['organizationId', 'caseId', 'clientId', 'branchId'] });
        console.log(`🚀 ~ module.exports.ClientVisitTask= ~ TaskResult:`, TaskResult)
        if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);

        let task = TaskResult.tasks.toJSON();

        return res.status(200).send({ status: true, task: task });
    }
    catch (e) {
        console.log("Error: ", e);
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

module.exports.ClientVisitTasks = async (req, res) => {
    try {
        const { _id: id, role } = req.local.user;

        const { organizationId, clientId, branchId, userId, start, end } = req.query;
        let query = { workType: "CLIENT_VISIT" };

        if (start && end) { query = { ...query, createdAt: dateRange(start, end) } };
        if (!!organizationId) { query = { ...query, organizationId: ObjectID(organizationId) } }
        if (!!clientId) { query = { ...query, clientId: ObjectID(clientId) } }
        if (!!branchId) { query = { ...query, branchId: ObjectID(branchId) } }
        if (!!userId) {
            query = { ...query, 'users.userId': ObjectID(userId) }
        }
        else if (['ADMIN', 'FRONTDESK'].includes(role)) {

        }
        else {
            query = { ...query, 'users.userId': ObjectID(id) }
        }

        const TaskResult = await TaskServices.findAll(query, { populate: ['organizationId', 'caseId', 'clientId', 'branchId'] }, req.local.startIndex, req.local.limit);
        if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);

        return res.status(200).send({ status: true, task: TaskResult.tasks, count: TaskResult.count });
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.ClientVisitTasks= ~ e:`, e.toString());
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

module.exports.ClientVisitCreateTask = async (req, res) => {
    try {
        const user = req.local.user.toJSON();
        const data = req.local.task;

        const taskData = {
            workType: workType.CLIENT_VISIT,
            organizationId: user.organization._id,
            users: [{ ...user, userId: user._id }],
            geo: { type: 'Point', coordinates: [data.location.longitude || 0, data.location.latitude || 0] },
            clientType: data.clientType,
            clientName: data.clientName,
            ...(data.clientId ? { clientId: data.clientId } : {}),
            ...(data.branchId ? { branchId: data.branchId } : {}),
            timeLog: [{ time: new Date(), type: "START", geo: { type: 'Point', coordinates: [data.location.longitude || 0, data.location.latitude || 0] } }],
            status: "MY_TASK"
        }

        const TaskResult = await TaskServices.create(taskData);
        if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);

        const newTaskId = TaskResult.tasks._id.toString();

        const NewTaskResult = await TaskServices.findById(newTaskId, { populate: ['organizationId', 'caseId', 'clientId', 'branchId'] });
        if (!NewTaskResult.status || NewTaskResult.error) return res.status(500).send(NewTaskResult);

        return res.status(200).send(NewTaskResult);
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.ClientVisitCreateTask= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
};

module.exports.ClientVisitUpdateTask = async (req, res) => {
    try {
        const user = req.local.user.toJSON();
        const taskId = req.params.id;
        const data = req.local.task;

        // get current task
        const currTaskRes = await TaskServices.findById(taskId);
        if (!currTaskRes.status || !currTaskRes.tasks._id) return res.status(500).send({ ...currTaskRes, error: "Task Not a found. Contact Administrator" })

        const targetFolderPath = path.join(process.env.CDN_PATH, 'client-visit', taskId);
        fs.mkdirSync(targetFolderPath, { recursive: true });

        // if getting checked out then take selfie and end the task
        if (data.action === "CHECK_OUT") {
            // if no selfie then throw an error
            if (req.files && req.files.images.length > 0) {
                await Promise.all(req.files.images.map(async file => {
                    const currentPath = path.join(process.cwd(), file.path);
                    const fileName = 'img_' + file.filename;
                    const destination = path.join(targetFolderPath, fileName);

                    await utilitiesController.moveFile(currentPath, destination);

                    const filepath = path.join('client-visit', taskId, fileName);
                    currTaskRes.tasks.images.push({
                        ...file,
                        fileName: fileName,
                        path: filepath,
                        type: "selfie"
                    })
                }));
            }
            else {
                return res.send({ status: false, error: "Selfie Required." })
            }

            currTaskRes.tasks.status = "COMPLETED";
            currTaskRes.tasks.timeLog.push({
                time: new Date(),
                type: "END",
                geo: { type: 'Point', coordinates: [data.location.longitude || 0, data.location.latitude || 0] }
            });
        }
        let coData = {
            creditOfficerId: null,
            firstName: "-",
            lastName: "-",
            role: "-",
        };

        if (data.creditOfficerId) {
            // find credit officer 
            const creditOfficer = await creditOfficerServices.findById(data.creditOfficerId);
            if (!creditOfficer.status || creditOfficer.error) return res.status(500).send(creditOfficer);
            coData = {
                creditOfficerId: creditOfficer.data._id,
                firstName: creditOfficer.data.firstName || "",
                lastName: creditOfficer.data.lastName || "",
                role: creditOfficer.data.role || "",
            };
        }

        if (data.audioNotes && data.audioNotes.length > 0) {

            const targetDirectory = path.join('client-visit', taskId);

            await Promise.all(data.audioNotes.map(async (base64) => {

                const fileName = 'audio_' + moment(new Date()).format("dd_mm_yyyy_hh_mm_ss") + ".m4a";
                const { status, filepath, error } = await base64toFile(base64, fileName, targetDirectory)
                if (status) {
                    currTaskRes.tasks.audioNotes.push({
                        id: Math.floor(100000 + Math.random() * 900000),
                        fileName: fileName,
                        path: filepath,
                        uploadedBy: req.local.user._id,
                        createdAt: new Date(),
                        ...coData
                    });
                }
            }));
        }

        if (data?.remark) {
            currTaskRes.tasks.remarks.push({
                id: Math.floor(100000 + Math.random() * 900000),
                remark: data.remark,
                uploadedBy: req.local.user._id,
                createdAt: new Date(),
                ...coData
            });
        }

        const result = await currTaskRes.tasks.save();
        return res.status(200).send({ status: true, data: result });
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.ClientVisitCreateTask= ~ err:`, err)
        return res.status(500).send({ status: false, error: err.toString() });
    }
};

module.exports.updateTaskWithCheck = async (query, data) => {
    try {
        const TaskResult = await TaskServices.updateByQuery(query, data);
        if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);

        return TaskResult
    }
    catch (err) {
        return { status: false, error: err };
    }
};

/**
 * @description This is used for update task
 * @param {*} req.params.id from that get taskId and from req.local.task get task Object
 * @param {*} res 
 * @returns Object task_res { status: true, tasks: object }
 * - { status : false, error : string }
 */
module.exports.pickTask = async (req, res) => {
    try {
        const taskId = req.params.id;
        const user = req.local.user.toJSON();
        const userId = user._id;
        if (!(taskId && taskId.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const taskPickedUpRes = await TaskServices.findAll({ _id: taskId, status: "OPEN" });
        if (taskPickedUpRes.tasks.length == 0) return res.status(500).send({ ...taskPickedUpRes, error: "Sorry you are bit late. Someone just picked up this task. Please Pick Next Task." })

        //  check count
        const totalTaskRes = await TaskServices.count({ 'users.userId': userId, status: 'MY_TASK' });
        if (!totalTaskRes.status || totalTaskRes.error) return res.status(500).send(totalTaskRes);

        const taskLimit = user.taskLimit || 50;
        if (totalTaskRes.count >= taskLimit) return res.status(500).send({ ...totalTaskRes, error: `You Have already ${taskLimit} Task in your task list. Please Complete the tasks and then you can pick another task.` })

        const userRes = await UserServices.findById(userId);
        if (!userRes.status || userRes.error) return res.status(500).send(userRes);

        const userObj = { ...userRes.users.toJSON(), userId };

        const TaskRes = await TaskServices.update(taskId, { $push: { users: [userObj] }, status: 'MY_TASK', pickedAt: new Date() });
        if (!TaskRes.status) return res.status(500).send(TaskRes);
        var taskResult = TaskRes.tasks.toJSON();

        let staff = {};

        switch (taskResult.workType) {
            case workType.RATE_VERIFY:
                staff = { "staff.rateVerifier": userRes.users };
                break;
            case workType.SITE_INSPECTION:
                staff = { "staff.siteEngineer": userRes.users };
                break;
            case workType.MAKER:
                staff = { "staff.maker": userRes.users };
                break;
            case workType.CHECKER:
                staff = { "staff.checker": userRes.users };
                break;
            case workType.DELIVERY:
                staff = { "staff.deliveryPerson": userRes.users };
                break;
            // case workType.SIGN:
            //     staff = { $push: { values: userRes.users } };
            //     break;
            default:
                break;
        }

        //  update the user of staff in case
        const caseRes = await CaseServices.updateCase(TaskRes.tasks.caseId, { $set: { ...staff } });
        return res.status(200).send({ status: true, tasks: taskResult });
    }
    catch (err) {
        logger.error("=> Task.controller = pickTask " + err.toString())
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description This is used for update task
 * @param {*} req.params.id from that get taskId and from req.local.task get task Object
 * @param {*} res 
 * @returns Object task_res { status: true, tasks: object }
 * - { status : false, error : string }
 */
module.exports.pickByCaseWorkType = async (req, res) => {
    try {
        const user = req.local.user.toJSON();
        const caseId = req.params.caseId;
        const workType = req.params.workType;
        const userId = user._id;
        if (!(caseId && workType)) return res.status(404).send({ status: false, error: "Not Found" });

        const taskPickedUpRes = await TaskServices.findAll({ caseId: caseId, workType: workType, status: "OPEN" });
        if (taskPickedUpRes.tasks.length == 0) return res.status(500).send({ ...taskPickedUpRes, error: "Sorry you are bit late. Someone just picked up this task. Please Pick Next Task." })

        const taskId = taskPickedUpRes.tasks[0]._id;
        //  check count
        const totalTaskRes = await TaskServices.count({ 'users.userId': userId, status: 'MY_TASK' });
        if (!totalTaskRes.status || totalTaskRes.error) return res.status(500).send(totalTaskRes);
        if (totalTaskRes.count >= 50) return res.status(500).send({ ...totalTaskRes, error: "You Have already 50 Task in your task list. Please Complete the tasks and then you can pick another task." })

        const userRes = await UserServices.findById(userId);
        if (!userRes.status || userRes.error) return res.status(500).send(userRes);

        const userObj = { ...userRes.users.toJSON(), userId };

        const TaskRes = await TaskServices.update(taskId, { $push: { users: [userObj] }, status: 'MY_TASK', pickedAt: new Date() });
        if (!TaskRes.status) return res.status(500).send(TaskRes);
        var taskResult = TaskRes.tasks.toJSON();

        let staff = {};

        switch (taskResult.workType) {
            case workType.SITE_INSPECTION:
                staff = { "staff.siteEngineer": userRes.users };
                break;
            case workType.MAKER:
                staff = { "staff.maker": userRes.users };
                break;
            case workType.CHECKER:
                staff = { "staff.checker": userRes.users };
                break;
            case workType.RATE_VERIFY:
                staff = { "staff.rateVerifier": userRes.users };
                break;
            case workType.DELIVERY:
                staff = { "staff.deliveryPerson": userRes.users };
                break;
            // case workType.SIGN:
            //     staff = { $push: { values: userRes.users } };
            //     break;
            default:
                break;
        }

        //  update the user of staff in case
        const caseRes = await CaseServices.updateCase(TaskRes.tasks.caseId, { $set: { ...staff } });
        return res.status(200).send({ status: true, tasks: taskResult });
    }
    catch (err) {
        console.log("🚀 ~ file: Task.controller.js:446 ~ module.exports.pickByCaseWorkType= ~ err:", err)
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description This is used for update task
 * @param {*} req.params.id from that get taskId and from req.local.task get task Object
 * @param {*} res 
 * @returns Object task_res { status: true, tasks: object }
 * - { status : false, error : string }
 */
module.exports.updateTask = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const checkTaskResult = await TaskServices.findAll({ _id: id }, {});
        if (!checkTaskResult.status || checkTaskResult.error) return res.send(checkTaskResult);
        if (checkTaskResult.count == 0) return res.send({ ...checkTaskResult, status: false });
        const taskRes = checkTaskResult.tasks[0];

        let data = req.local.task;

        if (data.userId) {
            const userRes = await UserServices.findById(data.userId);
            if (!userRes.status || userRes.error) return res.status(500).send(userRes);

            const userObj = { ...userRes.users.toJSON(), userId: data.userId };
            data = { ...data, users: [userObj], status: 'MY_TASK', pickedAt: new Date() };

            let staff = {};
            switch (taskRes.workType) {
                case workType.RATE_VERIFY:
                    staff = { "staff.rateVerifier": userRes.users };
                    break;
                case workType.SITE_INSPECTION:
                    staff = { "staff.siteEngineer": userRes.users };
                    break;
                case workType.MAKER:
                    staff = { "staff.maker": userRes.users };
                    break;
                case workType.CHECKER:
                    staff = { "staff.checker": userRes.users };
                    break;
                case workType.DELIVERY:
                    staff = { "staff.deliveryPerson": userRes.users };
                    break;
                default:
                    break;
            }

            //  update the user of staff in case
            const caseRes = await CaseServices.updateCase(taskRes.caseId, { $set: { ...staff } });

            delete data.userId;
        }

        const Task_res = await TaskServices.update(id, data);
        if (!Task_res.status) return res.status(500).send(Task_res);
        var taskResult = Task_res.tasks.toJSON();

        return res.status(200).send({ status: true, tasks: taskResult });
    }
    catch (err) {
        console.log("🚀 ~ file: Task.controller.js:197 ~ module.exports.updateTask= ~ err:", err)
        return res.status(500).send({ status: false, error: err });
    }
};

module.exports.reviewTask = async (req, res) => {
    try {
        const currTaskId = req.params.id;
        if (!(currTaskId && currTaskId.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });
        const tData = req.local.task;
        console.log("🚀 ~ file: Task.controller.js:334 ~ module.exports.reviewTask= ~ tData:", tData)

        const currTaskRes = await TaskServices.findById(currTaskId);
        if (!currTaskRes.status || !currTaskRes.tasks._id) return res.status(500).send({ ...currTaskRes, error: "Task Not a found. Contact Administrator" })
        const currTask = currTaskRes.tasks;
        const currTaskUser = (currTask.users.length > 0 && currTask.users.firstName) ? currTask.users[0] : {};

        const prevTaskRes = await TaskServices.findAll({ caseId: currTaskRes.tasks.caseId, workType: tData.workType });
        if (prevTaskRes.tasks.length == 0) return res.status(500).send({ ...prevTaskRes, error: "Task Not found. Contact Administrator" })
        const prevTaskId = prevTaskRes.tasks[0]._id;

        const TaskRes = await TaskServices.update(currTaskId, { status: 'REVIEW', $push: { review: { reason: tData.reason, toTask: prevTaskRes.tasks[0].toJSON(), date: new Date() } } });
        if (!TaskRes.status) return res.status(500).send(TaskRes);

        const TaskToRes = await TaskServices.update(prevTaskId, { status: 'MY_TASK', remark: `${currTaskUser.firstName + " " + currTaskUser.lastName}(${currTask.workType}): ${tData.reason}` });
        if (!TaskToRes.status) return res.status(500).send(TaskToRes);

        return res.status(200).send({ status: true, tasks: TaskToRes });
    }
    catch (err) {
        console.log("🚀 ~ file: Task.controller.js:197 ~ module.exports.updateTask= ~ err:", err)
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description This is used for update task
 * @param {*} req.params.id from that get taskId and from req.local.task get task Object
 * @param {*} res 
 * @returns Object task_res { status: true, tasks: object }
 * - { status : false, error : string }
 */
module.exports.updateTaskStatus = async (req, res) => {
    try {
        const user = req.local.user.toJSON();
        const taskId = req.params.id;
        const status = req.params.status;
        let caseUpdateData = {};

        if (!(taskId && taskId.length != 0)) return res.status(404).send({ status: false, error: "Task Id Not Found" });

        const TaskResult = await TaskServices.findById(taskId, { populate: ["caseId"] });
        if (!TaskResult.status || TaskResult.error) return res.status(500).send(TaskResult);
        const task = TaskResult.tasks;
        const caseId = task.caseId;
        let users = [];
        let note = "No Task Open for Other staff user.";

        // other task of same case
        const otherTasksResult = await TaskServices.findAll({ caseId: caseId._id });
        const otherTasks = otherTasksResult.tasks || [];

        if (status == "IN_PROGRESS") {

        }
        else if (status == "PAUSE") {

        }
        else if (task.status != "OPEN" && status == "MY_TASK") {
            // check for reason and store in case;
            const data = { $push: { 'remarks': { ...req.body, createdBy: user, createdAt: new Date(), } } };
            data.updatedBy = user._id;

            const case_res = await CaseServices.updateCase(caseId._id, data);
        }
        else if (status == "MY_TASK") {

        }
        else if (status == "COMPLETED") {
            switch (task.workType) {
                case workType.RATE_VERIFY:
                    if (["LEAD", "APPROXIMATE_LEAD"].indexOf(caseId.type) >= 0) { break; }
                    caseUpdateData = { caseStatus: "RATEVERIFICATIONCOMPLETED" };
                    users = (caseId.staff && caseId.staff.maker && caseId.staff.maker._id) ? [{ ...caseId.staff.maker, userId: caseId.staff.maker._id }] : [];
                    makerRTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.MAKER, users, status: users.length > 0 ? "MY_TASK" : "OPEN" });
                    break;
                case workType.SITE_INSPECTION:
                    caseUpdateData = { caseStatus: "SITEVISITCOMPLETED" };
                    users = (caseId.staff && caseId.staff.maker && caseId.staff.maker._id) ? [{ ...caseId.staff.maker, userId: caseId.staff.maker._id }] : [];
                    // console.log("🚀 ~ file: Task.controller.js:515 ~ module.exports.updateTaskStatus= ~ caseId.staff.maker:", caseId.staff.maker)
                    const makerTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.MAKER, users, status: users.length > 0 ? "MY_TASK" : "OPEN" });
                    break;
                case workType.MAKER:
                    users = (caseId.staff && caseId.staff.checker && caseId.staff.checker._id) ? [{ ...caseId.staff.checker, userId: caseId.staff.checker._id }] : [];
                    caseUpdateData = { $set: { caseStatus: "READYFORCHECKER", "staff.checker": users[0] } };
                    if (users.length == 0) {
                        const TaskResult = await TaskServices.findAll({ caseId: caseId._id, workType: workType.RATE_VERIFY }, { populate: ["caseId", "users.userId"] });
                        const task = TaskResult.tasks;
                        users = [{ ...task[0].users[0].userId.toJSON(), userId: task[0].users[0].userId._id }];
                        caseUpdateData = { $set: { caseStatus: "READYFORCHECKER", "staff.checker": users[0] } };
                    }
                    const checkerTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.CHECKER, users, status: users.length > 0 ? "MY_TASK" : "OPEN" });
                    break;
                case workType.CHECKER:
                    caseUpdateData = { caseStatus: "REPORTGENERATED" };
                    const prepareInvoiceTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.PREPARE_INVOICE });
                    if (prepareInvoiceTaskRes.status) {
                        const invRes = await nInvoiceController.createInvoiceRequestSystem({ caseId, user })
                    }
                    break;
                case workType.PREPARE_INVOICE:
                    const hasCheckerTask = _.find(otherTasks, ['workType', 'CHECKER']);
                    // if checker task complete the update status and create print task
                    if (hasCheckerTask && hasCheckerTask.status == "COMPLETED") {
                        caseUpdateData = { caseStatus: "READYFORPRINT" };
                        const printTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.PRINT });
                    }
                    break;
                case workType.PRINT:
                    caseUpdateData = { caseStatus: "ATTACHSIGN" };
                    // const signTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.SIGN });
                    const signedReportUploadTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.SIGNED_REPORT_UPLOAD });
                    break;
                case workType.SIGN:
                    // const signedReportUploadTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.SIGNED_REPORT_UPLOAD });
                    break;
                case workType.SIGNED_REPORT_UPLOAD:
                    caseUpdateData = { caseStatus: "READYFORDELIVERY" };
                    const deliveryTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.DELIVERY });
                    break;
                case workType.DELIVERY:
                    caseUpdateData = { caseStatus: "DELIVERED" };
                    const collectOcCopyTaskRes = await this.createTask({ caseId: caseId._id, organizationId: task.organizationId, workType: workType.COLLECT_OC_COPY });
                    break;
                case workType.PAYMENT_COLLECT:
                    caseUpdateData = { caseStatus: "COMPLETED" };
                    break;
                default:
                    break;
            }

            if (caseId.type == "LEAD") {
                delete caseUpdateData.caseStatus
            }

            if (Object.keys(caseUpdateData).length > 0) {
                // update caseStatus based on task status update;
                const doneCaseRes = await CaseServices.updateCase(caseId._id, caseUpdateData);
            }

            const reviewTasks = await TaskServices.findAll({ caseId: caseId._id, 'review.toTask._id': taskId, status: 'REVIEW' }, { populate: ['caseId', 'users.userId'] });

            if (!!reviewTasks.status && reviewTasks.tasks.length > 0) {
                const reviewTask = reviewTasks.tasks[0];
                const TaskToRes = await TaskServices.update(reviewTask._id, { status: 'MY_TASK' });
                console.log("Reviewed reopen task: ", TaskToRes);
                note = "Task reopened for " + (reviewTask.users[0].userId.firstName || "") + " " + (reviewTask.users[0].userId.lastName || "");

                const notificationToken = reviewTask.users[0].userId.notificationToken || null;
                const sent = await sendExpoNotification([notificationToken], `Review ${reviewTask.workType.replace('_', ' ')}  Task is Competed (Case Number ${caseId.caseNumber}). You can Start working on it now.`, {})

                console.log("🚀 ~ file: Task.controller.js:458 ~ module.exports.updateTaskStatus= ~ notificationToken:", reviewTask.users[0].userId, notificationToken)
            } else {
                console.log("no review tasks")
            }
        }

        const Task_res = await TaskServices.update(taskId, { status, completedAt: new Date() });

        if (!Task_res.status) return res.status(500).send(Task_res);
        var taskResult = Task_res.tasks.toJSON();

        return res.status(200).send({ status: true, tasks: taskResult.tasks, note });
    }
    catch (err) {
        logger.error("🚀 ~ file: Task.controller.js:252 ~ module.exports.updateTaskStatus= ~ err:" + err.toString())
        return res.status(500).send({ status: false, error: err });
    }
};

// /**
//  *
//  * @param {*} req
//  * @param {*} res
//  * @returns Object { status: true, message: "Task Deleted" }
//  *
//  */
// module.exports.deleteTask = async (req, res) => {
//     try {
//         const taskId = req.params.id;

//         const task = await TaskServices.deleteTask(taskId);
//         if (!task.status) return res.status(500).send(task);

//         return res.status(200).send({ status: true, message: "Task Deleted" });
//     }
//     catch (e) {
//         console.log(e);
//         return res.status(500).send({ status: false, error: e });
//     }
// }