const fs = require('fs');
const path = require('path');

const CONFIG = require('../Constants/Config');
const Zone = require('../Model/Zone');

const ZoneServices = require('../Services/Zone.services');
const ClientServices = require('../Services/Client.services');
const BranchServices = require('../Services/Branch.services');
const CaseServices = require('../Services/Case.services');
const OrganizationServices = require('../Services/Organization.services');
const EmpanelmentServices = require('../Services/CreateServices')('Empanelment');

const MailController = require('../Controllers/Mail.controller');
const utilitiesController = require('../Controllers/Utilities.controller');
const { logger } = require('../logger');
const { ObjectID } = require('mongodb');

/**
 * @description This is used for get list of all zones
 * @param {*} req 
 * @param {*} res 
 * @returns Object ZoneResult { status: true, zones: Array }
 * - { status : false, error : string }
 */
module.exports.getAllZones = async (req, res) => {
    try {
        let query = {
            // 'organization._id': req.local.user.organization._id;
        };
        var count = 0;
        var zoneRes = [];

        const { searchValue, client, zone } = req.query;

        if (searchValue) {
            query['$or'] = [
                { "client.name": { $regex: searchValue, $options: 'i' } },
                { zoneName: { $regex: searchValue, $options: 'i' } },
                { "client.initial": { $regex: searchValue, $options: 'i' } }
            ]
        }

        // For Advanced Search By ID
        if (client) { query['client._id'] = client }
        if (zone) { query['_id'] = zone }

        const ZoneResult = await ZoneServices.getAllZones(query, req.local.limit, req.local.startIndex);
        if (!ZoneResult.status || ZoneResult.error) return res.status(500).send(ZoneResult);

        await Promise.all(ZoneResult.zones.map(async (zone) => {
            var zone = zone.toJSON();
            const EmpanelmentResult = await EmpanelmentServices.find({ query: { zoneId: zone._id, isDelete: false }, limit: 1000, populate: ['zoneId', 'valuerId'] });
            // console.log(EmpanelmentResult)
            zone.Empanelments = EmpanelmentResult.data;
            zoneRes.push(zone);

        }))

        const ZoneResultCount = await ZoneServices.getCount(query);
        if (!ZoneResultCount.status || ZoneResultCount.error) return res.status(500).send(ZoneResultCount);

        count = ZoneResultCount.zones;
        return res.status(200).send({ status: true, zones: zoneRes, count: count });
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.getAllZones= ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};
/**
 * @description This is used for get one zone
 * @param {*} req.params.id from that get zoneId  
 * @param {*} res 
 * @returns Object ZoneResult { status: true, ZoneResult, Empanelments: Object.values(arr) }
 * - { status : false, error : string }
 */
module.exports.getZone = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const ZoneResult = await ZoneServices.getZone(id);
        if (!ZoneResult.status || ZoneResult.error) return res.status(500).send(ZoneResult);

        const { masterTypeOfPurpose, assetsClass } = req.query;
        if (masterTypeOfPurpose && assetsClass) {

            const EmpanelmentResult = await EmpanelmentServices.find({ query: { zoneId: id, masterTypeOfPurpose: masterTypeOfPurpose, assetsClass: assetsClass, isDelete: false }, limit: 1000, populate: ['zoneId', 'valuerId'] });
            if (!EmpanelmentResult.status || EmpanelmentResult.error) return res.status(500).send(EmpanelmentResult);

            return res.status(200).send({ status: true, ZoneResult, Empanelments: EmpanelmentResult.data });

        }
        const EmpanelmentResult = await EmpanelmentServices.find({ query: { zoneId: id, isDelete: false }, limit: 1000, populate: ['zoneId', 'valuerId'] });
        if (!EmpanelmentResult.status || EmpanelmentResult.error) return res.status(500).send(EmpanelmentResult);

        return res.status(200).send({ status: true, ZoneResult, Empanelments: EmpanelmentResult.data });
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.getZone= ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

module.exports.getOneZone = async (req, res) => {
    try {
        const id = req.params.zone_id;

        const ZoneResult = await Zone.findOne({ _id: id }, { _id: 1, zoneName: 1 })
            .then((Zone) => {
                return { status: true, zones: Zone };
            }).catch((e) => {
                return { status: false, error: e };
            });

        return res.status(200).send(ZoneResult);
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.getOneZone= ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};
/**
 * @description This is used for create new zone
 * @param {*} req.local.zone from that get zone Object
 * @param {*} res 
 * @returns Object ZoneResult { status: true, zones: object }
 * - { status : false, error : string }
 */
module.exports.createZone = async (req, res) => {
    try {
        const datas = req.local.zone;
        const orgId = req.local.user.organization;

        const Organizations = await OrganizationServices.getById(orgId);
        if (!Organizations.status || Organizations.error) return res.status(500).send(Organizations);
        datas.organization = Organizations.Organizations;

        if (datas.client) {

            const Client = await ClientServices.getClientById(datas.client);
            if (!Client.status) return res.status(500).send(Client);

            datas.client = Client.clients
        }
        const ZoneResult = await ZoneServices.createZone(datas);
        if (!ZoneResult.status || ZoneResult.error) return res.status(500).send(ZoneResult);

        if (datas.empanelments) {
            var zone = ZoneResult.zones.toJSON();
            var arr = [];
            const id = zone._id.toString();
            await Promise.all(datas.empanelments.map(async (emp) => {

                if (!fs.existsSync(process.env.CDN_PATH + '/empanelmentLetters')) {
                    fs.mkdirSync(process.env.CDN_PATH + '/empanelmentLetters');
                }

                emp.zoneId = id;
                const EmpanelmentResult = await EmpanelmentServices.create(emp);

                // Moving File
                if (!fs.existsSync(process.env.CDN_PATH + '/empanelmentLetters/' + id)) {
                    fs.mkdirSync(process.env.CDN_PATH + '/empanelmentLetters/' + id);
                }
                const empLetter = [];
                await Promise.all(EmpanelmentResult.data.empanelmentLetter.map(async file => {

                    file.fileName = "empanelmentLetters_" + file.fileName;
                    let currentPath = process.cwd() + "/" + file.path;
                    const destination = path.join(process.env.CDN_PATH, "empanelmentLetters", id, file.fileName);

                    // call function for move file 
                    await utilitiesController.moveFile(currentPath, destination);

                    file.path = path.join("empanelmentLetters", id, file.fileName);
                    await empLetter.push(file);

                }))

                EmpanelmentResult.data.empanelmentLetter = empLetter;
                await EmpanelmentResult.data.save();
                arr.push(EmpanelmentResult.data);

            }))
            // console.log("Zone Create Response : ", JSON.stringify(zone));
            return res.status(200).send({ status: true, zones: zone });
        }
        return res.status(200).send(ZoneResult);
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.createZone= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
};
/**
 * @description This is used for update zone
 * @param {*} req.params.id from that get zoneId and from req.local.zone get zone Object
 * @param {*} res 
 * @returns Object zone_res { status: true, zones: object }
 * - { status : false, error : string }
 */
module.exports.updateZone = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const datas = req.local.zone;
        const checkEmpanelExist = await EmpanelmentServices.find({ query: { "zoneId": id }, limit: 1000, populate: ['zoneId', 'valuerId'] });
        if (!checkEmpanelExist.status || checkEmpanelExist.error) return res.status(500).send(checkEmpanelExist);

        if (datas.client) {

            const Client = await ClientServices.getClientById(datas.client);
            if (!Client.status) return res.status(500).send(Client);

            datas.client = Client.clients
        }
        const Zone_res = await ZoneServices.updateZone(id, datas);
        if (!Zone_res.status) return res.status(500).send(Zone_res);
        var zoneResult = Zone_res.zones.toJSON();

        zoneResult.empanelments = checkEmpanelExist.data;

        if (datas.empanelments) {

            var arr = [];
            await Promise.all(datas.empanelments.map(async (emp, index) => {
                emp.zoneId = zoneResult._id;

                // If new Empanelment Create

                if (!emp._id) {
                    const EmpanelmentResult = await EmpanelmentServices.create(emp);

                    // // Moving File
                    // if (!fs.existsSync(process.env.CDN_PATH + '/empanelmentLetters/' + id)) {
                    //     fs.mkdirSync(process.env.CDN_PATH + '/empanelmentLetters/' + id);
                    // }
                    // const empLetter = [];
                    // await Promise.all(EmpanelmentResult.Empanelment.empanelmentLetter.map(async file => {

                    //     file.fileName = "empanelmentLetters_" + file.fileName;
                    //     let currentPath = process.cwd() + "/" + file.path;
                    //     const destination = path.join(process.env.CDN_PATH, "empanelmentLetters", id, file.fileName);

                    //     // call function for move file 
                    //     await utilitiesController.moveFile(currentPath, destination);

                    //     file.path = path.join("empanelmentLetters", id, file.fileName);
                    //     await empLetter.push(file);

                    // }))

                    // EmpanelmentResult.Empanelment.empanelmentLetter = empLetter;
                    // await EmpanelmentResult.Empanelment.save();
                    arr.push(EmpanelmentResult.data);
                }
            }))
            zoneResult.empanelments = arr;
            return res.status(200).send({ status: true, zones: zoneResult });
        }
        // console.log(zoneResult)
        return res.status(200).send({ status: true, zones: zoneResult });
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.updateZone= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
};
/**
 * @description This is used for get list of all branches Zone wise 
 * @param {*} req.params.id from that get zoneId
 * @param {*} res 
 * @returns Object BranchResult { status: true, branches: branches }
 * - { status : false, error : string }
 */
module.exports.getAllBranchesByZone = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const BranchResult = await BranchServices.getAllBranches({ "zone._id": id });
        if (!BranchResult.status || BranchResult.error) return res.status(500).send(BranchResult);

        return res.status(200).send(BranchResult);
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.getAllBranchesByZone= ~ e:` + e.toString())
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};


//=========================== This is Incomplete Function ============================
module.exports.calculation = async (req, res) => {
    try {
        const datas = req.body;
        // console.log(datas);
        var amount = datas.amount;
        var fees;
        await datas.feeStructure.map(data => {
            if (amount > datas.from && amount < datas.to) {
                fees += parseInt(datas.value);
                amount = amount - datas.to;
                if (datas.to == "") {

                }
            }
        })
        // return res.status(200).send(BranchResult);
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.calculation= ~ e:` + e.toString())
        res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * 
 * @param {*} req 
 * @param {*} res 
 * @returns Object { status: true, message: "Zone Deleted" }
 * 
 */
module.exports.deleteZone = async (req, res) => {
    try {
        const zoneId = req.params.id;

        const branch_res = await BranchServices.getAllBranches({ 'zone._id': zoneId });
        if (!branch_res.status) return res.status(500).send(branch_res);

        if (branch_res.branches.length == 0) {

            const zone = await ZoneServices.deleteZone(zoneId);
            if (!zone.status) return res.status(500).send(zone);

            const EmpanelmentResult = await EmpanelmentServices.updateMany({ zoneId: zoneId }, { isDelete: true });
            if (!EmpanelmentResult.status || EmpanelmentResult.error) return res.status(500).send(EmpanelmentResult);

            return res.status(200).send({ status: true, message: "Zone Deleted" });

        }
        return res.status(500).send({ status: false, error: "Zone can't Delete. Because zone assigned in Branch " });
    }
    catch (e) {
        logger.error(`🚀 ~ module.exports.deleteZone ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: e.toString() });
    }
}
// =================================== Empanelment Function ====================================
/**
 * @description This function is used for check Empanelment already exist or not 
 * @param {*} req.body.empNo from that get empanelment No 
 * @param {*} res 
 * @returns Object { status: true }
 * - { status: false, error: "Already Created!" }
 */
module.exports.checkEmpanelmentNoExist = async (req, res) => {
    try {
        const empNo = req.body.empNo.trim();

        const empanelCheck = await EmpanelmentServices.find({ query: { empanelmentNo: empNo, isDelete: false }, limit: 1000, populate: ['zoneId', 'valuerId'] })
        if (!empanelCheck.status) return res.status(500).send({ status: false, empanelCheck });

        if (empanelCheck.data.length == 0) {

            return res.status(200).send({ status: true });
        }
        else {
            return res.status(500).send({ status: false, error: "Already Created!" });
        }
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.checkEmpanelmentNoExist= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
}

module.exports.getEmpanelments = async (req, res) => {
    try {
        let { searchText = "", clientId, zoneId, valuerId, assetsClass } = req.query;

        let commonQuery = [
            {
                '$match': {
                    isDelete: false,
                    ...(assetsClass && { assetsClass: assetsClass }),
                    ...(zoneId && { zoneId: ObjectID(zoneId) }),
                    ...(valuerId && { valuerId: ObjectID(valuerId) })
                }
            },
            {
                '$lookup': {
                    from: "zones",
                    localField: "zoneId",
                    foreignField: "_id",
                    as: "zoneId",
                }
            },
            {
                '$addFields': { "zoneId": { $first: "$zoneId" } }
            },
        ];
        const empanelRes = await EmpanelmentServices.aggregate([
            ...commonQuery,
            {
                '$lookup': {
                    from: "users",
                    localField: "valuerId",
                    foreignField: "_id",
                    as: "valuerId",
                }
            },
            {
                '$match': {
                    ...(clientId && { 'zoneId.client._id': clientId }), // Conditional matching based on clientId existence
                    ...(!!searchText && { // Conditional matching based on searchText existence
                        $or: [
                            { 'empanelmentNo': { $regex: searchText, $options: 'i' } }, // Matching empanelmentNo with searchText
                            { 'valuerId.firstName': { $regex: searchText, $options: 'i' } }, // Matching firstName with searchText
                            { 'valuerId.lastName': { $regex: searchText, $options: 'i' } }, // Matching lastName with searchText
                            { 'zoneId.name': { $regex: searchText, $options: 'i' } }, // Matching zoneId name with searchText
                        ]
                    })
                }
            },
            { '$skip': req.local.startIndex },
            { '$limit': req.local.limit },
        ]);
        if (!empanelRes.status) return res.status(500).send({ status: false, empanelRes });

        const empanelCountRes = await EmpanelmentServices.aggregate([
            ...commonQuery,
            {
                '$match': { ...(clientId && { 'zoneId.client._id': clientId }) }
            },
            { '$count': 'count' }
        ]);

        const EmpanelmentCount = (empanelCountRes.data.length > 0 && empanelCountRes.data[0].count) ? empanelCountRes.data[0].count : 0

        return res.status(200).send({ status: true, data: empanelRes.data, count: EmpanelmentCount });
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.getEmpamelments= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
}

/**
 * @description This function is used for get Empanelment details By Id
 * @param {*} req 
 * @param {*} res 
 * @returns Object EmpanelmentResult  { status: true , Empanelment : Object}
 * - { status: false, err }
 */
module.exports.getEmpanelment = async (req, res) => {
    try {
        const id = req.params.id;

        const EmpanelmentResult = await EmpanelmentServices.findById(id);
        if (!EmpanelmentResult.status) return res.status(500).send({ status: false, EmpanelmentResult });

        return res.status(200).send({ status: true, Empanelment: EmpanelmentResult.data });
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.getEmpanelment= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
}
/**
 * @description This function is used for update Empanelment details 
 * @param {*} req.params.id from that get empanelment id, from req.body get update empanelment details Object 
 * @param {*} res 
 * @returns Object EmpanelmentResult  { status: true , Empanelment : Object}
 * - { status: false, err }
 */
module.exports.updateEmpanelment = async (req, res) => {
    try {
        const id = req.params.id;
        const data = req.body;

        const EmpanelmentResult = await EmpanelmentServices.updateById(id, data);
        if (!EmpanelmentResult.status) return res.status(500).send({ status: false, EmpanelmentResult });

        return res.status(200).send({ status: true, Empanelment: EmpanelmentResult.data });
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.updateEmpanelment= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
}
/**
 * @description This function is used for delete Empanelmet Letter
 * @param {*} req.body.empNo from that get empanelment No 
 * @param {*} res 
 * @returns Object { status: true, message: "Empanelment Letter Deleted" }
 * -{ status: false, error: "Empanelment Letter can't delete. Because Empanelment Letter attached in case. " }
 */

module.exports.deleteEmpanelment = async (req, res) => {
    try {
        const empanelmentNumber = req.body.empNo;
        const id = req.params.id;

        const findEmpanelInCase = await CaseServices.getAllCases({ empanelmentNo: { $in: [empanelmentNumber] } });
        if (!findEmpanelInCase.status) return res.status(500).send(findEmpanelInCase);

        if (findEmpanelInCase.cases != 0) {

            const findEmpanelmentInCase = await CaseServices.getAllCases({ caseStatus: { $in: ["CANCELLED", "CLOSED"] }, empanelmentNo: { $in: [empanelmentNumber] } });

            if (!findEmpanelmentInCase.status) return res.status(500).send(findEmpanelmentInCase);

            if (findEmpanelmentInCase.cases.length != 0) {

                const EmpanelmentResult = await EmpanelmentServices.updateById(id, { isDelete: true });
                if (!EmpanelmentResult.status) return res.status(500).send({ status: true, Empanelment: EmpanelmentResult.data });

                return res.status(200).send({ status: true, message: "Empanelment Letter Deleted" });
            }
            else {
                return res.status(500).send({ status: false, error: "Empanelment Letter can't delete. Because Empanelment Letter attached in case. " });
            }
        }

        const EmpanelmentResult = await EmpanelmentServices.updateById(id, { isDelete: true });
        if (!EmpanelmentResult.status) return res.status(500).send({ status: true, Empanelment: EmpanelmentResult.data });

        return res.status(200).send({ status: true, message: "Empanelment Letter Deleted" });
    }

    catch (e) {
        logger.error(`🚀 ~ module.exports.deleteEmpanelment ~ e:` + e.toString())
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
}
/**
 * @description This function is used for share empanelment letter with common Email
 * @param {*} req
 * @param {*} res
 * @returns
 */
module.exports.shareEmpanelment = async (req, res) => {
    try {
        const id = req.params.id;
        const email = req.body.email;
        const mailSubject = req.body.subject;
        const mailBody = req.body.mailBody;

        const EmpanelmentResult = await EmpanelmentServices.findById(id);
        if (!EmpanelmentResult.data) return res.status(500).send({ status: false, EmpanelmentResult });

        const attachmentsArr = await Promise.all(EmpanelmentResult.data.empanelmentLetter.map(async (emp) => {

            if (EmpanelmentResult.data.fileType == "Pdf") {

                let parts = emp.replace(/\//g, '').split("\\");
                const fileName = parts[parts.length - 1]; // Or parts.pop();

                return { fileName: fileName, path: CONFIG.STAGINGURL + emp.replace(/\//g, '') };
            }
            return { fileName: emp.fileName, path: CONFIG.STAGINGURL + emp.path };
        }))

        const html = "";
        const subject = mailSubject;
        const from = "Vastukala";
        const to = email;
        const message = mailBody;
        const attachments = attachmentsArr;

        const mailResult = await MailController.SendMail(from, to, subject, message, attachments, html);
        return res.status(200).send(mailResult);
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.shareEmpanelment= ~ err:` + err.toString())
        return res.status(500).send({ status: false, error: err.toString() });
    }
}