const fs = require('fs');
const path = require('path');
const _ = require('loadsh');

const CONFIG = require('../Constants/Config');
const Invoice = require('../Model/Invoice');
const Task = require("../Model/Task");
const { workType } = require('../Types/Task');

const TaskController = require('./Task.controller');
const UtilityController = require('./Utilities.controller');

const CaseServices = require('../Services/Case.services');
const nInvoiceServices = require('../Services/CreateServices')('nInvoice');
const { logger } = require('../logger');
const { dateRange } = require('../Utils/Utility');
const { ObjectID } = require('mongodb');

/**
 * @description This is used for get list of all payments
 * @param {*} req
 * @param {*} res
 * @returns Object PaymentResult { status: true, payments: Array }
 * - { status : false, error : string }
 */
module.exports.getDashboardInvoiceCount = async (req, res) => {
    try {
        let responseCounts = {
            OPEN: { count: 0 },
            PENDING: { count: 0 },
            COMPLETED: { count: 0 },
            CANCELLED: { count: 0 }
        };

        const InvoiceCount = await nInvoiceServices.aggregate([
            { $group: { _id: "$invoiceStatus", count: { $sum: 1 } } },
        ]);
        if (!InvoiceCount.status || InvoiceCount.error) return res.status(500).send(InvoiceCount);

        const statusCounts = InvoiceCount.data.map((st, id) => { responseCounts[st._id].count = st.count; });

        return res.status(200).send({ status: true, counts: responseCounts });
    }
    catch (e) {
        logger.error(`🚀 ~ file: Invoice.controller.js:54 ~ module.exports.getDashboardInvoiceCount= ~ e: ${e.toString()}`)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used for get list of all invoices
 * @param {*} req 
 * @param {*} res 
 * @returns Object InvoiceResult { status: true, invoices: Array }
 * - { status : false, error : string }
 */
module.exports.getAllInvoices = async (req, res) => {
    try {
        const { invoiceStatus, searchText, start, end } = req.query;
        let query = { ...(invoiceStatus && { invoiceStatus: invoiceStatus }) };

        if (invoiceStatus == "COMPLETED") { query = { ...query, updatedAt: dateRange(start, end) } }

        if (searchText) {
            query = {
                ...query,
                '$or': [
                    { 'cases.caseNumber': { $regex: searchText, $options: 'i' } },
                    { 'client.name': { $regex: searchText, $options: 'i' } },
                    { 'branch.name': { $regex: searchText, $options: 'i' } },
                    { 'cases.ownerName': { $regex: searchText, $options: 'i' } },
                    { 'invoiceNo': { $regex: searchText, $options: 'i' } }
                ]
            };
        }

        const InvoiceResult = await nInvoiceServices.find({ query, limit: req.local.limit, start: req.local.startIndex });
        if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);

        return res.status(200).send({ status: true, invoices: InvoiceResult.data, count: InvoiceResult.count });
    }
    catch (e) {
        logger.error(`🚀 ~ file: nInvoice.controller.js:68 ~ module.exports.getAllInvoices= ~ e:, ${e.toString()}`)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used for get one invoice
 * @param {*} req.params.id from that get invoiceId  
 * @param {*} res 
 * @returns Object InvoiceResult { status: true, InvoiceResult, Empanelments: Object.values(arr) }
 * - { status : false, error : string }
 */
module.exports.getInvoice = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Invalid Request!" });

        // const InvoiceResult = await nInvoiceServices.findById(id);
        const InvoiceResult = await nInvoiceServices.aggregate([
            { '$match': { _id: ObjectID(id) } },
            {
                '$lookup': {
                    from: 'cases',
                    localField: 'cases._id',
                    foreignField: '_id',
                    as: 'casesInfo'
                }
            }
        ]);
        if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);
        if (InvoiceResult.data.length == 0) return res.status(500).send({ status: false, message: "Invoice not found!" });

        let finalInvoiceResult = InvoiceResult.data[0]
        finalInvoiceResult.casesInfo.map((csi, cid) => {
            let caseIndex = finalInvoiceResult.cases.findIndex(cs => cs._id.toString() === csi._id.toString());
            finalInvoiceResult.cases[caseIndex] = { ...csi, ...finalInvoiceResult.cases[caseIndex] };
        });
        delete finalInvoiceResult.casesInfo;

        if (finalInvoiceResult.invoiceType === "TAX") {
            const proformaInvoiceResult = await nInvoiceServices.find({ query: { taxInvoiceId: ObjectID(id) } });
            if(proformaInvoiceResult.count > 0) { 
                finalInvoiceResult.pgInvoice = proformaInvoiceResult.data[0]
            }
        }

        return res.status(200).send({ status: true, data: finalInvoiceResult });
    }
    catch (e) {
        logger.error(`🚀 ~ file: nInvoice.controller.js:56 ~ module.exports.getInvoice= ~ e:, ${e.toString()}`)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

/**
 * @description This is used for create new invoice
 * @param {*} req.local.invoice from that get invoice Object
 * @param {*} res 
 * @returns Object InvoiceResult { status: true, invoices: object }
 * - { status : false, error : string }
 */
module.exports.createInvoiceRequest = async (req, res) => {
    try {
        const user = req.local.user;
        const data = req.local.invoice;
        if (!data.cases) { data.cases = [data.caseId] }

        const CaseResult = await CaseServices.find({ query: { _id: { $in: data.cases } } });
        if (!CaseResult.status || CaseResult.error) return res.status(500).send(CaseResult);
        if (CaseResult.cases.length <= 0) { return res.status(404).send({ status: false, error: "Case Not Found!" }) }
        const casesData = CaseResult.cases.map((cs, id) => { return { ...cs.toJSON(), fees: cs.toJSON().fees.expectedPrice || 0 } });
        const firstCase = casesData[0];

        const organization = firstCase.organization;
        const gstNo = firstCase.gstNo || "";

        const invoice = {
            note: data.note || "SYSTEM GENERATE WHEN Lead/CASE CREATED",
            invoiceType: data.invoiceType || "PROFORMA",
            isAdvance: true,
            amount: data.amount || 0,
            pendingAmount: data.amount || 0,
            requestedBy: user,
            cases: casesData,
            organization: organization,
            gstNo: gstNo,
        };

        const advanceInvoiceTaskRes = await TaskController.createTask({ organizationId: organization._id, caseId: data.caseId, workType: workType.PREPARE_INVOICE });

        const InvoiceResult = await nInvoiceServices.create(invoice);
        if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);
        return res.status(200).send(InvoiceResult);
    }
    catch (err) {
        logger.error("nInvoice error request: " + err.toString())
        return res.status(500).send({ status: false, error: err });
    }
};

/**
 * @description This is used for create new invoice request from attach report/report generate
 * @param {*} data from that get invoice Object
 * @returns Object InvoiceResult { status: true, invoices: object }
 * - { status : false, error : string }
 */
module.exports.createInvoiceRequestSystem = async (data) => {
    try {
        const user = data.user;
        data.organization = { name: user.organization.name, id: user.organization._id };

        const CaseResult = await CaseServices.findCase(data.caseId);
        if (!CaseResult.status || CaseResult.error) return res.status(500).send(CaseResult);
        if (!CaseResult.cases) { return res.status(404).send({ status: false, error: "Case Not Found!" }) }
        const caseData = CaseResult.cases.toJSON();

        const hasAmount = parseInt((caseData.fees && caseData.fees.quotedAmount) ? caseData.fees.quotedAmount : 0)
        const amount = hasAmount ? hasAmount : 0

        const hasInvoiceResult = await nInvoiceServices.find({ query: { 'cases._id': data.caseId, invoiceStatus: "OPEN" } });
        logger.error(`🚀 ~ module.exports.createInvoiceRequestSystem= ~ 176 line amount:` + hasAmount + "");
        if (!hasInvoiceResult.status || hasInvoiceResult.error) return res.status(500).send(hasInvoiceResult);

        if (hasInvoiceResult.count == 0) {
            const createInvoiceResult = await nInvoiceServices.create({
                note: "SYSTEM GENERATE WHEN CHECKER Checked CREATED",
                invoiceType: "PROFORMA",
                amount: amount,
                pendingAmount: amount,
                cases: [{ ...caseData, fees: caseData.fees.expectedPrice || 0 }],
                clientType: caseData.clientType,
                client: caseData.client,
                branch: caseData.branch || null,
                organization: caseData.organization,
                requestedBy: user,
                requestedAt: new Date(),
                gstNo: caseData.gstNo || "",
            });

            if (!createInvoiceResult.status || createInvoiceResult.error) return { ...createInvoiceResult };
            return createInvoiceResult;

        } else {
            return { status: false, error: "Open invoices found." };
        }
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.createInvoiceRequestSystem= ~ err:` + err.toString())
        return { status: false, error: err };
    }
};

/**
 * @description This is used for update invoice
 * @param {*} req.params.id from that get invoiceId and from req.local.invoice get invoice Object
 * @param {*} res 
 * @returns Object invoice_res { status: true, invoices: object }
 * - { status : false, error : string }
 */
module.exports.updateInvoice = async (req, res) => {
    try {
        const id = req.params.id;
        if (!(id && id.length != 0)) return res.status(404).send({ status: false, error: "Not Found" });

        const { user, invoice: invoiceBody } = req.local;

        const InvoiceResult = await nInvoiceServices.findById(id);
        if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);

        const CaseResult = await CaseServices.find({ query: { _id: { $in: invoiceBody.cases.map((c, id) => c._id) } }, limit: 1000 });
        if (!CaseResult.status || CaseResult.error) return res.status(500).send(CaseResult);
        if (!CaseResult.cases) { return res.status(404).send({ status: false, error: "Case Not Found!" }) }

        invoiceBody.cases = CaseResult.cases.map((c, id) => {
            const cData = c.toJSON();
            const _case = invoiceBody.cases.find((cd, id) => cd._id == cData._id) || {};
            return { ...cData, fees: _case.fees || 0, pendingAmount: _case.fees || 0 }
        });

        if (!req.file) {
            return res.status(500).send({ status: false, error: "No Invoice File Found" });
        }

        if (!fs.existsSync(process.env.CDN_PATH + '/invoices/' + id)) {
            fs.mkdirSync(process.env.CDN_PATH + '/invoices/' + id, { recursive: true });
        }
        const fileName = "inv_" + req.file.filename;
        const destination = path.join(process.env.CDN_PATH, "invoices", id, fileName);

        // call function for move file
        let currentPath = process.cwd() + "/" + req.file.path;
        await UtilityController.moveFile(currentPath, destination);
        const filepath = path.join("invoices", id, fileName);

        let newData = {
            ...invoiceBody,
            invoiceStatus: "COMPLETED",
            completedAt: new Date(),
            completedBy: user,
            invoicedAmount: (+invoiceBody.amount - (+invoiceBody.receivedAmount || 0)) || 0,
            pendingAmount: (+invoiceBody.amount - (+invoiceBody.receivedAmount || 0)) || 0,
            invoiceFile: {
                ...req.file,
                fileName,
                path: filepath,
                documentType: { key: "INVOICE", value: "INVOICE" },
                uploadedBy: user,
                uploadedAt: new Date()
            }
        };

        const updateInvoiceRes = await nInvoiceServices.updateById(id, { $set: { ...newData } });
        if (!updateInvoiceRes.status) return res.status(500).send(updateInvoiceRes);

        return res.status(200).send({ status: true, invoices: {} });
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports.updateInvoice= ~ err: ` + err.toString());
        return res.status(500).send({ status: false, error: err.toString() });
    }
};

// /**
//  * @param {*} req 
//  * @param {*} res 
//  * @returns Object { status: true, message: "Invoice Deleted" }
//  */
// module.exports.deleteInvoice = async (req, res) => {
//     try {
//         const invoiceId = req.params.id;

//         const InvoiceResult = await nInvoiceServices.deleteById(invoiceId);
//         if (!InvoiceResult.status) return res.status(500).send(InvoiceResult);

//         return res.status(200).send({ status: true });
//     }
//     catch (e) {
//         logger.error(`🚀 ~ file: nInvoice.controller.js:274 ~ module.exports.deleteInvoice ~ e:, ${e.toString()}`)
//         return res.status(500).send({ status: false, error: e });
//     }
// }

/**
 * @param {*} req 
 * @param {*} res 
 * @returns Object { status: true, message: "Invoice Updated" }
 */
module.exports.cancelInvoice = async (req, res) => {
    try {
        const invoiceId = req.params.id;

        const InvoiceResult = await nInvoiceServices.updateById(invoiceId, { invoiceStatus: "CANCELLED" });
        if (!InvoiceResult.status) return res.status(500).send(InvoiceResult);

        let newInv = InvoiceResult.data.toJSON();
        delete newInv._id;
        delete newInv.createdAt;
        delete newInv.updateAt;
        delete newInv.updatedBy;
        delete newInv.invoiceFile;
        newInv.pendingAmount = +newInv.amount - (+newInv.receivedAmount || 0) || 0

        const newInvResult = await nInvoiceServices.create({ ...newInv, invoiceStatus: "OPEN" });
        if (!newInvResult.status) return res.status(500).send(newInvResult);
        const newInvoice = newInvResult.data;

        return res.status(200).send({ status: true, newInvoice });
    }
    catch (e) {
        logger.error(`🚀 ~ file: nInvoice.controller.js:274 ~ module.exports.deleteInvoice ~ e:, ${e.toString()}`)
        return res.status(500).send({ status: false, error: e.toString() });
    }
}

/**
 * @param {*} req 
 * @param {*} res 
 * @returns Object { status: true, message: "Invoice Deleted" }
 */
module.exports.deleteInvoice = async (req, res) => {
    try {
        const invoiceId = req.params.id;
        const { user } = req.local;

        const InvoiceResult = await nInvoiceServices.updateById(invoiceId, { invoiceStatus: "CANCELLED", note: "deleted", cancelledAt: new Date(), cancelledBy: user });
        if (!InvoiceResult.status) return res.status(500).send(InvoiceResult);

        return res.status(200).send({ status: true });
    }
    catch (e) {
        logger.error(`🚀 ~ file: nInvoice.controller.js:274 ~ module.exports.deleteInvoice ~ e:, ${e.toString()}`)
        return res.status(500).send({ status: false, error: e.toString() });
    }
}

/**
 * @description This is used for get list of all invoices
 * @param {*} req 
 * @param {*} res 
 * @returns Object InvoiceResult { status: true, invoices: Array }
 * - { status : false, error : string }
 */
module.exports.mergeInvoice = async (req, res) => {
    try {
        const { invoiceNo } = req.query;

        const InvoiceResult = await nInvoiceServices.aggregate([
            { '$match': { invoiceNo, invoiceStatus: { $nin: ["CANCELLED", "CANCELED"] } } },
            {
                '$group': {
                    _id: '$invoiceNo',
                    cases: {
                        $addToSet: { $first: "$cases" }
                    },
                    invoiceIds: { $addToSet: '$_id' },
                    total: { $sum: 1 }
                }
            }
        ]);
        if (!InvoiceResult.status || InvoiceResult.error) return res.status(500).send(InvoiceResult);
        console.log(`🚀 ~ file: nInvoice.controller.js:441 ~ module.exports.mergeInvoice= ~ InvoiceResult:`, InvoiceResult)

        if (InvoiceResult.data.length <= 0) return res.status(200).send({ status: true, message: 'no invoice found' })
        const inv = InvoiceResult.data[0];
        const [invoiceId, ...deleteIds] = inv.invoiceIds;
        console.log(`🚀 ~ file: nInvoice.controller.js:446 ~ module.exports.mergeInvoice= ~ invoiceId, deleteIds:`, invoiceId, deleteIds)

        const updateResult = await nInvoiceServices.updateById(inv.invoiceIds[0], [{ $set: { cases: inv.cases } }]);
        const deletedResult = await nInvoiceServices.deleteByQuery({ _id: { $in: deleteIds } })
        return res.status(200).send({ status: true, updateResult, deletedResult });
    }
    catch (e) {
        logger.error(`🚀 ~ file: nInvoice.controller.js:68 ~ module.exports.getAllInvoices= ~ e:, ${e.toString()}`)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};

module.exports.invoicesByCase = async (req, res) => {
    try {
        const { caseId } = req.params;
        let query = {};


        const caseInvoiceResult = await CaseServices.aggregate([
            {
                $match: {
                    _id: ObjectID(caseId)
                }
            },
            {
                "$lookup": {
                    from: "ninvoices",
                    localField: "_id",
                    foreignField: "cases._id",
                    as: "invoices",
                }
            },
            {
                "$project": {
                    invoices: 1,
                    ownerName: 1,
                    caseNumber: 1,
                    fullAddress: 1,
                    caseStatus: 1,
                    paymentStatus: 1,

                }
            }
        ]);
        if (!caseInvoiceResult.status || caseInvoiceResult.error) return res.status(500).send(caseInvoiceResult);
        if (caseInvoiceResult.data.length == 0) return res.status(500).send({ status: false, error: "Not Found" });

        return res.status(200).send({ status: true, data: caseInvoiceResult.data[0] });
    }
    catch (e) {
        logger.error(`🚀 ~ file: nInvoice.controller.js:432 ~ module.exports.invoicesByCase= ~ e:, ${e.toString()}`)
        return res.status(500).send({ status: false, error: "Internal Server Error!" });
    }
};