const Invoice = require('../Model/Invoice');
const defaultUserSelect = 'firstName lastName email profilePhoto';
const defaultPopulate = [
    { path: 'requestedBy', select: defaultUserSelect },
    { path: 'collectedBy', select: defaultUserSelect },
    { path: 'processedBy', select: defaultUserSelect },
]

/**
 * @description This is used for get All cases
 * @param {*} query 
 * @param {*} displayFields 
 * @param {*} startIndex 
 * @param {*} limit 
 * @returns Object { status: true, cases: Array }
 *  { status: false, error: String }
 */
module.exports.getAllInvoices = async (query = {}, { display = [], sort = { createdAt: -1 }, populate = [] } = {}, limit, startIndex) => {

    return Invoice.find(query, display).populate([...defaultPopulate, ...populate]).sort(sort).skip(startIndex).limit(limit)
        .then((invoices) => {
            return { status: true, invoices: invoices };
        }).catch((e) => {
            return { status: false, error: e };
        });
};

/**
 * @description This is used for get One invoice By Id 
 * @param {String} id 
 * @returns Object { status: true, invoices: Object }
 * - { status: false, error: string }
 */
module.exports.getInvoice = async (id) => {

    return Invoice.findById(id).then((invoices) => {
        return { status: true, invoices: invoices };
    }).catch((e) => {
        return { status: false, error: e };
    });
};

module.exports.findOne = async (query) => {

    return Invoice.findOne(query).then((invoices) => {
        return { status: true, invoice: invoices };
    }).catch((e) => {
        return { status: false, error: e };
    });
};

/**
 * @description This is used for  create Invoice 
 * @param {*} InvoiceData 
 * @returns  Object { status: true, invoices: Object }
 * - { status: false, error: string }
 */
module.exports.createInvoice = async (InvoiceData) => {

    return Invoice.create(InvoiceData)
        .then((result) => { return { status: true, invoices: result }; })
        .catch((err) => {
            console.log(err)
            return { status: false, error: err }
        })
};

/**
 * @description  This is used for update Invoice 
 * @param {String} id 
 * @param {Object} InvoiceData 
 * @returns  Object { status: true, invoices: Object }
 * - { status: false, error: string }
 */
module.exports.updateInvoice = async (id, InvoiceData) => {

    return Invoice.findOneAndUpdate({ _id: id }, InvoiceData, { new: true })
        .then((result) => { return { status: true, invoices: result }; })
        .catch((err) => {
            console.log(err)
            return { status: false, error: err }
        })
};

/**
 * @description This is used for update invoice from query
 * @param {Object} find 
 * @param {Object} InvoiceData 
 * @returns Object { status: true, invoices: Object }
 * - { status: false, error: string }
 */
module.exports.findUpdateInvoice = async (find, InvoiceData) => {

    return Invoice.findOneAndUpdate(find, InvoiceData, { new: true })
        .then((result) => { return { status: true, invoices: result }; })
        .catch((err) => {
            console.log("🚀 ~ file: Invoice.services.js:80 ~ module.exports.findUpdateInvoice= ~ err:", err)
            return { status: false, error: err }
        })
};

/**
 * @description This is used for update Many invoices
 * @param {Object} find_query 
 * @param {Object}  InvoiceData
 * @returns Object { status: true, invoices: result }
 * -  { status: false, err }
 */
module.exports.updateMultipleInvoice = async (id, update, filter) => {
    return Invoice.updateMany(id, update, filter)
        .then((result) => { return { status: true, invoices: result } })
        .catch((err) => {
            console.log(err)
            return { status: false, err }
        })
};

/**
 * @description  This is used for delete Invoice 
 * @param {String} id 
 * @returns  Object { status: true, invoices: Object }
 * - { status: false, error: string }
 */
module.exports.deleteInvoice = async (id) => {
    return Invoice.remove({ _id: id })
        .then((result) => { return { status: true, invoices: result }; })
        .catch((err) => {
            console.log(err)
            return { status: false, error: err }
        })
};
