const { ToWords } = require("to-words");
const moment = require("moment");
const json2xls = require('json2xls');
const path = require("path");
const fs = require("fs");
const { logger } = require("../logger");

exports.isObjectBlank = (object) => {
    for (const property in object) { return false; }
    return true;
}

exports.ConvertIntoRupees = (amount = 0, options = {}) => {
    try {
        const toWords = new ToWords();

        let words = toWords.convert(amount, { currency: true, ...options });
        return words;
    }
    catch (err) {
        logger.error("=> utility convertIntoRupees" + amount + " _ " + err.toString());
        return "-";
    }
}

exports.CarpetToBuiltUpConvert = (area = 0, builtUp_percentage = 0, isReverse = false) => {

    let totalAmount = 0;

    if (!!area && !!builtUp_percentage) {
        if (isReverse) {
            const builtUp_percentage_Area = ((parseFloat(area) * parseInt(builtUp_percentage)) / (100 + parseInt(builtUp_percentage)));
            totalAmount = (parseFloat(area) - parseFloat(builtUp_percentage_Area)).toFixed(2);
        } else {
            const builtUp_percentage_Area = (parseFloat(area) * parseFloat(builtUp_percentage)) / 100;
            totalAmount = (parseFloat(area) + parseFloat(builtUp_percentage_Area)).toFixed(2);
        }
    }
    return totalAmount;
}

exports.SquareFeetToMeter = (area = 0, isReverse = false) => {
    let covertArea = 0;
    if (isReverse) {
        // in meter to feet
        covertArea = parseFloat(parseFloat(area).toFixed(2) / 10.7639).toFixed(2);
    } else {
        // in feet to meter
        covertArea = parseFloat(parseFloat(area).toFixed(2) * 10.7639).toFixed(2);
    }
    return covertArea;
}

module.exports.deepFlattenObject = (arrayOfObjects, prefix = '') => {

    let result = {};
    function recurse(current, flattenedKey) {
        for (let key in current) {
            let newKey = flattenedKey ? `${flattenedKey}_${key}` : key;
            if (key == "_id") {
                result[newKey] = "" + current[key];
            }
            else if (Array.isArray(current[key])) {
                current[key].forEach((item, index) => { recurse(item, `${newKey}_${index}`); });
            } else if (typeof current[key] === 'object' && current[key] !== null) {
                recurse(current[key], newKey);
            } else {
                result[newKey] = current[key];
            }
        }
    }

    return arrayOfObjects.map((obj, index) => {
        result = {}
        recurse(obj, prefix);
        return result;
    });
}

module.exports.exportToCsv = async (results) => {
    const headers = [];
    for (const result of results) {
        Object.keys(result).forEach((key) => {
            if (!headers.includes(key)) {
                headers.push(key);
            }
        });
    }
    let rows = "";
    rows += headers.join(',') + '\n';

    for (const result of results) {
        const row = new Array(headers.length).map((item) => '');

        // Match obj.key to same index in header
        Object.keys(result).forEach((key) => {
            const keyPosInHeader = headers.indexOf(key);
            const processedResult =
                typeof result[key] === 'object'
                    ? /**
                 * Handling for fields containing objects including nested objects?
                 * Just replace commas with colons for now */
                    JSON.stringify(result[key]).replaceAll(',', ';')
                    : JSON.stringify(result[key]);

            // Match item to existing header columns
            if (keyPosInHeader !== -1) {
                // JSON.stringify here since fields can be ObjectId or Object
                // Both are not compatible for conversion to string directly
                row[keyPosInHeader] = processedResult;
            } else {
                row.push(processedResult);
            }
        });
        rows += row.join(',') + ',\n';
    }
    return rows
};

module.exports.dateRange = (startDate, endDate, format = "DD-MM-YYYY") => {

    const isValidDate = (dateString) => moment(dateString, format).isValid();

    if (!startDate || !endDate || !isValidDate(startDate) || !isValidDate(endDate)) {
        const currentDate = Date.now();
        endDate = new Date(moment(currentDate).endOf('month'));
        startDate = new Date(moment(currentDate).startOf('month'));
    } else {
        startDate = new Date(moment(startDate, format).startOf('day'));
        endDate = new Date(moment(endDate, format).endOf('day'));
    }

    return { $gte: startDate, $lte: endDate }
}

module.exports.generateExcel = async (data = [{}], fileName = "download.xlsx") => {
    try {

        const nData = data.map((item, index) => {
            // Incremental SrNo starting from 1
            return Object.assign({ SrNo: index + 1 }, item);
        });
        const xls = json2xls(nData);
        const filePath = path.join(process.env.CDN_FULL_PATH, "temp");
        if (!fs.existsSync(filePath)) {
            fs.mkdirSync(filePath, { recursive: true });
        }

        fs.writeFileSync(`${filePath}/${fileName}`, xls, 'binary');

        return `${filePath}/${fileName}`;
    }
    catch (e) {
        throw new Error(e.toString())
    }
}