const { workTypeArr, workType } = require("../Types/Task");
const TaskController = require('../Controllers/Task.controller');
const _ = require('loadsh');
const { logger } = require('../logger');

const Task = require("../Model/Task");
const Case = require("../Model/Case");
const User = require("../Model/User");
const { sendExpoNotification } = require("../Services/ExpoNotification.services");
const TaskLog = require("../Model/TaskLog");
const CaseAuditLog = require("../Model/CaseAuditLog");

const getNewKey = (str) => str.replace(/\./gi, '_');

const replaceAllObjKeys = (obj, getNewKey) => {
    if (Array.isArray(obj)) {
        for (let i = 0; i < obj.length; i++) {
            replaceAllObjKeys(obj[i], getNewKey);
        }
    }
    else if (typeof obj === "object") {
        for (const key in obj) {
            const newKey = getNewKey(key);
            obj[newKey] = obj[key];
            if (key !== newKey) {
                delete obj[key];
            }
            replaceAllObjKeys(obj[newKey], getNewKey);
        }
    }
    return obj;
};

module.exports = async (change) => {

    const collection = change.ns.coll;
    const operationType = change.operationType;
    const id = change.documentKey._id || null;
    const data = (operationType == "update") ? change.updateDescription.updatedFields : change.fullDocument;
    const nData = replaceAllObjKeys(_.cloneDeep(data), getNewKey);

    try {
        if (collection == "cases") {
            const caseAuditLogRes = await CaseAuditLog.create({ caseId: id, data: nData, operationType, updatedBy: data.updatedBy });
            const caseRes = await Case.findById(id);

            if (operationType == "delete") {
                // deleted task when case is deleted
                const taskRes = await Task.updateMany({ caseId: id, status: { $ne: 'COMPLETED' } }, { status: 'CANCELLED' });
            }
            else if (operationType == "update") {
                if (!!data.organization && data.organization._id) {
                    const taskRes = await Task.updateMany({ caseId: id, organizationId: { $ne: data.organization._id } }, { organizationId: data.organization._id });
                }
                if (!!data.location) {
                    const caseUpdateRes = await Case.updateOne({ _id: id }, { geo: { type: 'Point', coordinates: [data.location.longitude || 0, data.location.latitude || 0] } })
                    const siteInspectionTaskUpdateRes = await Task.updateOne({ caseId: id }, { geo: { type: 'Point', coordinates: [data.location.longitude || 0, data.location.latitude || 0] } });
                }
                //  update tasks if user assigned at case update
                if (!!data.staff && !!data.staff.rateVerifier) {
                    const rateVerifierTaskRes = await Task.updateOne({ workType: workType.RATE_VERIFY, caseId: id, 'users.userId': { $ne: data.staff.rateVerifier._id }, status: { $nin: ['COMPLETED', 'CANCELLED'] } }, { users: [{ ...data.staff.rateVerifier, userId: data.staff.rateVerifier._id }], status: "MY_TASK" });
                }
                if (!!data.staff && !!data.staff.siteEngineer) {
                    const siteInspectionTaskRes = await Task.updateOne({ workType: workType.SITE_INSPECTION, caseId: id, 'users.userId': { $ne: data.staff.siteEngineer._id }, status: { $nin: ['COMPLETED', 'CANCELLED'] } }, { users: [{ ...data.staff.siteEngineer, userId: data.staff.siteEngineer._id }], status: "MY_TASK" });
                }
                if (!!data.staff && !!data.staff.maker) {
                    const makerTaskRes = await Task.updateOne({ workType: workType.MAKER, caseId: id, 'users.userId': { $ne: data.staff.maker._id }, status: { $nin: ['COMPLETED', 'CANCELLED'] } }, { users: [{ ...data.staff.maker, userId: data.staff.maker._id }], status: "MY_TASK" });
                }
                if (!!data.staff && !!data.staff.checker) {
                    const checkerTaskRes = await Task.updateOne({ workType: workType.CHECKER, caseId: id, 'users.userId': { $ne: data.staff.checker._id }, status: { $nin: ['COMPLETED', 'CANCELLED'] } }, { users: [{ ...data.staff.checker, userId: data.staff.checker._id }], status: "MY_TASK" });
                }
                // temporary for old system
                if (!!data.caseStatus && ['SITEVISITCOMPLETED', 'READYFORCHECKER', 'SIGNEDREPORTATTACH', "ATTACHSIGN", 'DELIVERED', 'READYFORPRINT'].includes(data.caseStatus)) {
                    const siteInspectionTaskRes = await Task.updateOne({ workType: workType.SITE_INSPECTION, caseId: id, status: { $nin: ['COMPLETED', 'CANCELLED'] } }, { status: "COMPLETED", completedAt: new Date() });
                }
                if (data.caseStatus == "DELIVERED") {
                    const taskRes = await Task.updateMany({ caseId: id, workType: { $in: ["MAKER", "CHECKER", "SITE_INSPECTION", "DELIVERY", "RATE_VERIFY"] }, status: { $ne: 'COMPLETED' } }, { status: 'COMPLETED' });
                }
                if (data.caseStatus == "CANCELLED") {
                    const taskRes = await Task.updateMany({ caseId: id, workType: { $nin: ["LEAD", "CASE"] }, status: { $ne: 'COMPLETED' } }, { status: 'CANCELLED', canceledAt: new Date() });
                }
                if (data.caseStatus == "RE_OPENED") {
                    const myTaskTaskRes = await Task.updateMany({ caseId: id, 'users.0': { $exists: true }, status: { $eq: 'CANCELLED' } }, { status: 'MY_TASK', canceledAt: new Date() });
                    const openTaskRes = await Task.updateMany({ caseId: id, 'users.0': { $exists: false }, status: { $eq: 'CANCELLED' } }, { status: 'OPEN', canceledAt: new Date() });
                }
            }
            else if (operationType == "insert") {

                if (data.type == "CASE") {
                    // new case so = CREATE TASK RATE_VERIFY AND SITE_INSPECTION
                    const pickupPerson = caseRes.staff.pickupPerson;
                    const pickedUpTaskRes = await TaskController.createTask({ caseId: id, organizationId: data.organization._id, workType: workType.CASE, users: [{ ...pickupPerson, userId: pickupPerson._id }], status: 'COMPLETED', completedAt: new Date() });

                    const rateVerifyPerson = (!!caseRes.staff && !!caseRes.staff.rateVerifier && !!caseRes.staff.rateVerifier._id) ? { status: 'MY_TASK', users: [{ ...data.staff.rateVerifier, userId: data.staff.rateVerifier._id }] } : {}
                    const rateVerifyTaskRes = await TaskController.createTask({ caseId: id, organizationId: data.organization._id, workType: workType.RATE_VERIFY, ...rateVerifyPerson });

                    const siteEngineerPerson = (!!caseRes.staff && !!caseRes.staff.siteEngineer && !!caseRes.staff.siteEngineer._id) ? { status: 'MY_TASK', users: [{ ...data.staff.siteEngineer, userId: data.staff.siteEngineer._id }] } : {}
                    const siteInspectionTaskRes = await TaskController.createTask({ caseId: id, organizationId: data.organization._id, geo: { type: 'Point', coordinates: [data.location.longitude || 0, data.location.latitude || 0] }, workType: workType.SITE_INSPECTION, ...siteEngineerPerson });

                    if (!!data.staff && !!data.staff.maker && !!data.staff.maker._id) {
                        const makerTaskRes = await TaskController.createTask({ workType: workType.MAKER, caseId: id, organizationId: data.organization._id, users: [{ ...data.staff.maker, userId: data.staff.maker._id }], status: "MY_TASK" });
                    }
                    if (!!data.staff && !!data.staff.checker && !!data.staff.checker._id) {
                        const checkerTaskRes = await TaskController.createTask({ workType: workType.CHECKER, caseId: id, organizationId: data.organization._id, users: [{ ...data.staff.checker, userId: data.staff.checker._id }], status: "MY_TASK" });
                    }

                } else if (data.type == "APPROXIMATE_LEAD") {
                    const rateVerifyTaskRes = await TaskController.createTask({ caseId: id, organizationId: data.organization._id, workType: workType.RATE_VERIFY });
                }
            }
        }
        else if (collection == "tasks") {
            const taskRes = await Task.findById(id).populate(['caseId', 'users.userId']);
            const taskId = taskRes._id;
            const caseId = taskRes.caseId || null;

            if (!data.timeLog) {
                const taskLogRes = await TaskLog.create({ taskId, workType: taskRes.workType, data: nData, operationType });
            }

            if (operationType == "update") {
                // send Notification 
                if (taskRes.status == "COMPLETED") {
                    // const reviewTasks = await Task.find({ caseId: caseId, 'review.toTask._id': taskId }).populate(['caseId', 'users.userId']);
                    // const reviewTask = reviewTasks.length > 0 ? reviewTasks[0] : null;
                    // if (!!reviewTask && reviewTask.users.length > 0 && reviewTask.users[0].notificationToken) {
                    //     const sent = await sendExpoNotification([reviewTask.users[0].notificationToken], `Review ${reviewTask.workType.replace('_', ' ')}  Task is Competed (Case Number ${caseId.caseNumber}). You can Start working on it now.`, { taskRes })
                    // };
                }
                if (taskRes.status == "MY_TASK") {
                    if (taskRes.users.length > 0 && taskRes.users[0].notificationToken) {
                        const sent = await sendExpoNotification([taskRes.users[0].notificationToken], `${taskRes.workType.replace('_', ' ')} Task Assigned to you. Please check it.`, { type: 'TASK', taskId: taskId })
                    }
                }
            }
            else if (operationType == "insert" && taskRes.status == "OPEN") {
                // find all the users which ha this workType permission
                // send notification to all the users
                let notificationTokens = await User.aggregate([
                    {
                        '$match': {
                            'type': { '$ne': 'CREDITOFFICER' },
                            'role': { '$exists': true },
                            'isdelete': false,
                            'notificationToken': { '$exists': true }
                        }
                    }, {
                        '$lookup': {
                            'from': 'roles',
                            'localField': 'role',
                            'foreignField': 'name',
                            'as': 'roleDetails'
                        }
                    },// { '$unwind': '$roleDetails' },
                    {
                        $match: {
                            "roleDetails.taskWorkTypes": {
                                $elemMatch: { $in: [taskRes.workType, "$roleDetails.taskWorkTypes"] },
                            },
                        }
                    },
                    { $project: { notificationToken: '$notificationToken' } }
                ]);
                notificationTokens = notificationTokens.reduce((prev, curr) => { return [...prev, curr.notificationToken] }, [])
                const sent = await sendExpoNotification(notificationTokens, `New ${taskRes.workType.replace('_', ' ')} Task Created, Hurry Up to pick it.`, { taskId })
            }
            else {

            }
        }
    }
    catch (err) {
        logger.error(`🚀 ~ module.exports= ~ err:`, err.toString())
        // return UserUpdates.create({ data: change, db: db, collectionName: 'users', status: 'failed', err: JSON.stringify(err), type: 'error_log' });
    }
};